/*
** File: platform.c  
** Project: ADMXRC3 API module library
** Purpose: Implements Linux-specific functions for ADMXRC3 API module library
**
** (C) Copyright Alpha Data 2010
*/

#include "platform.h"

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>

#include "../../../platform/linux/ioctl_dflinux.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

static const char g_deviceName[] = "admxrc3";

/* Unmap the first mmap'ed region, as part of cleanup-on-close */
static ADMXRC3_STATUS
unmapFirstRegion(
  ADMXRC3_HANDLE hAdmxrc3)
{
  DfLinuxIoctlFindMmapping io;
  int err, ioCode;

  io.in.bByIndex = TRUE;
  io.in.variant.index = 0U;
  ioCode = _IOC(_IOC_READ | _IOC_WRITE, 150, DFLINUX_IOCTLCODE_FINDMMAPPING, 0);
  err = ioctl(hAdmxrc3, ioCode, &io);
  if (0 != err) {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      default:
        fprintf(stderr, "unmapFirstRegion: ioctl() failed, errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      switch ((DfLinuxMmapStatus)err) {
      case DfLinuxMmapInvalidRegion:
        return ADMXRC3_INVALID_REGION;

      case DfLinuxMmapInvalidTag:
        return ADMXRC3_INVALID_INDEX;

      case DfLinuxMmapNoMemory:
        return ADMXRC3_NO_MEMORY;

      case DfLinuxMmapRegionTooLarge:
        return ADMXRC3_REGION_TOO_LARGE;

      case DfLinuxMmapAccessDenied:
        return ADMXRC3_ACCESS_DENIED;

      case DfLinuxMmapGeneralFailure:
      default:
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    }
  }
  if (0 == munmap((void*)(uintptr_t)io.out.virtualBase, (size_t)io.out.virtualLength)) {
    return ADMXRC3_SUCCESS;
  } else {
      switch (errno) {
      case EINVAL:
        return ADMXRC3_INVALID_REGION;

      default:
        fprintf(stderr, "unmapFirstRegion: munmap() failed, errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
  }
}

ADMXRC3_STATUS
admxrc3OpenDevice(
  unsigned int index,
  bool_t bUnprivileged,
  ADMXRC3_HANDLE* phAdmxrc3)
{
  char filename[32];
  int hDevice;

  snprintf(filename, ARRAY_LENGTH(filename), "/dev/%s%u", g_deviceName, index);
  hDevice = open(filename, bUnprivileged ? O_RDONLY : O_RDWR);
  if (hDevice < 0) {
    switch (errno) {
    case EACCES:
      return ADMXRC3_ACCESS_DENIED;

    case ENXIO:
    case ENOENT:
      return ADMXRC3_DEVICE_NOT_FOUND;

    default:
      fprintf(stderr, "admxrc3OpenDevice: errno=%d (%s)\n", errno, strerror(errno));
      return ADMXRC3_UNEXPECTED_ERROR;
    }
  }
  *phAdmxrc3 = hDevice;
  return ADMXRC3_SUCCESS;
}

ADMXRC3_STATUS
admxrc3CloseDevice(
  ADMXRC3_HANDLE hAdmxrc3)
{
  ADMXRC3_STATUS status;
  int hAdmxrc3Duped;

  /* Duplicate the device handle (temporarily). */
  hAdmxrc3Duped = dup(hAdmxrc3);
  if (hAdmxrc3Duped < 0) {
    switch (errno) {
    case EBADF:
      return ADMXRC3_INVALID_HANDLE;

    default:
      /* Failed to duplicate the device handle; fall back to other method of cleaning up mmap'ed regions */
      break;
    }

    /* Attempt to munmap() any existing mappings by repeatedly munmap'ing the first mapping */
    while (1) {
      status =  unmapFirstRegion(hAdmxrc3);
      if (ADMXRC3_SUCCESS != status) {
        break;
      }
    }
    
    if (close(hAdmxrc3) < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;
	
      default:
        fprintf(stderr, "admxrc3CloseDevice: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    }    
  } else {
    /* Close the device handle. This prevents client code from using the handle to do further operations (such as mmap) */
    if (close(hAdmxrc3) < 0) {
      switch (errno) {
      case EBADF:
        status = ADMXRC3_INVALID_HANDLE;
        break;
	
      default:
        fprintf(stderr, "admxrc3CloseDevice: errno=%d (%s)\n", errno, strerror(errno));
        status = ADMXRC3_UNEXPECTED_ERROR;
        break;
      }
    }

    /* Attempt to unmap any existing mappings by repeatedly munmap'ing the first mapping */
    while (1) {
      status =  unmapFirstRegion(hAdmxrc3Duped);
      if (ADMXRC3_SUCCESS != status) {
        break;
      }
    }
    
    /* Now close the duplicated (temporary) device handle */
    close(hAdmxrc3Duped);
  }

  return ADMXRC3_SUCCESS;
}

#define MAX_IOCTL_SIZE (64)
#define IOCTL_SIZE_IN_SHIFT  (0)
#define IOCTL_SIZE_OUT_SHIFT (6)

ADMXRC3_STATUS
admxrc3Ioctl(
  ADMXRC3_HANDLE hAdmxrc3,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  if (sizeIn >= MAX_IOCTL_SIZE || sizeOut >= MAX_IOCTL_SIZE) {
    fprintf(stderr, "admxrc3Ioctl: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)code, ((int)sizeIn << IOCTL_SIZE_IN_SHIFT) | ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdmxrc3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        fprintf(stderr, "admxrc3Ioctl: EINPROGRESS returned for blocking operation\n");
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3Ioctl: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3IoctlNonBlock(
  ADMXRC3_HANDLE hAdmxrc3,
  ADMXRC3_TICKET* pTicket, /* ignored */
  unsigned int code,
  void* pIoctl,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  if (sizeIn >= MAX_IOCTL_SIZE || sizeOut >= MAX_IOCTL_SIZE) {
    fprintf(stderr, "admxrc3IoctlNonBlock: sizeIn (%u) or sizeOut (%u) greater than 64\n", sizeIn, sizeOut);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)code | 0x40U, ((int)sizeIn << IOCTL_SIZE_IN_SHIFT) | ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdmxrc3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3IoctlNonBlock: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3FinishNonBlock(
  ADMXRC3_HANDLE hAdmxrc3,
  ADMXRC3_TICKET* pTicket,
  bool_t bWait,
  void* pIoctl,
  unsigned int sizeOut)
{
  static const int ioctlType = 150;
  int ioctlCode, err;
  struct pollfd fds[1];

  if (bWait) {
    fds[0].fd = hAdmxrc3;
    fds[0].events = POLLPRI;
    if (poll(fds, 1, -1 /* negative value => wait for ever */) > 0) {
      /* Success */
    } else {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case EINTR:
        return ADMXRC3_CANCELLED;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINVAL:
      case EFAULT:
      default:
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    }
  }
  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)DFLINUX_IOCTLCODE_FINISH, ((int)sizeOut << IOCTL_SIZE_OUT_SHIFT));
  err = ioctl(hAdmxrc3, ioctlCode, pIoctl);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case EPIPE:
        return ADMXRC3_NONBLOCK_IDLE;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3FinishNonBlock: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3IoctlCancel(
  ADMXRC3_HANDLE hAdmxrc3)
{
  static const int ioctlType = 150;
  int ioctlCode, err;

  ioctlCode = _IOC(_IOC_READ | _IOC_WRITE, ioctlType, (int)DFLINUX_IOCTLCODE_CANCEL, 0);
  err = ioctl(hAdmxrc3, ioctlCode, 0);
  if (0 == err) {
    return ADMXRC3_SUCCESS;
  } else {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      case ENOMEM:
        return ADMXRC3_NO_MEMORY;

      case EINPROGRESS:
        fprintf(stderr, "admxrc3IoctlCancel: EINPROGRESS returned for blocking operation\n");
        return ADMXRC3_PENDING;

      case ECANCELED:
        return ADMXRC3_CANCELLED;

      case ENOTTY:
      case EFAULT:
      case EINVAL:
      default:
        fprintf(stderr, "admxrc3IoctlCancel: errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      return (ADMXRC3_STATUS)err;
    }
  }
}

ADMXRC3_STATUS
admxrc3MapRegion(
  ADMXRC3_HANDLE hAdmxrc3,
  unsigned int windowIndex,
  uint64_t offset,
  uint64_t length,
  void** ppVirtualBase)
{
  DfLinuxIoctlGetMmapAddress io;
  uint8_t* p;
  int err, ioCode;
  long pageSize;
  uint64_t address;
  off_t pageMask, pageStart;
  size_t pageLength, pageOffset;

  assert(sizeof(off_t) == sizeof(unsigned long));

  /* Sanity checks */
  if (0 == length || offset + length < offset) {
    return ADMXRC3_INVALID_REGION;
  }

  ioCode = _IOC(_IOC_READ | _IOC_WRITE, 150, DFLINUX_IOCTLCODE_GETMMAPADDRESS, 0);
  io.in.regionTag = windowIndex;
  err = ioctl(hAdmxrc3, ioCode, &io);
  if (0 != err) {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      default:
        fprintf(stderr, "admxrc3MapRegion: ioctl() failed, errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      switch ((DfLinuxMmapStatus)err) {
      case DfLinuxMmapInvalidRegion:
        return ADMXRC3_INVALID_REGION;

      case DfLinuxMmapInvalidTag:
        return ADMXRC3_INVALID_INDEX;

      case DfLinuxMmapNoMemory:
        return ADMXRC3_NO_MEMORY;

      case DfLinuxMmapRegionTooLarge:
        return ADMXRC3_REGION_TOO_LARGE;

      case DfLinuxMmapAccessDenied:
        return ADMXRC3_ACCESS_DENIED;

      case DfLinuxMmapGeneralFailure:
      default:
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    }
  }
  pageSize = sysconf(_SC_PAGESIZE);
  if (pageSize == (unsigned long)-1) {
    fprintf(stderr, "admxrc3MapRegion: get page size from sysconf() failed, errno=%d (%s)\n", errno, strerror(errno));
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  pageMask = (off_t)(pageSize - 1U);
  address = io.out.address + offset;
  if (length + offset > io.out.length) {
    return ADMXRC3_INVALID_REGION;
  }
  pageStart = (off_t)address & ~pageMask;
  pageOffset = (size_t)address & (size_t)pageMask;
  pageLength = (size_t)((size_t)length + pageOffset + (size_t)pageMask) & ~(size_t)pageMask;
  p = (uint8_t*)mmap(NULL, pageLength, PROT_READ | PROT_WRITE, MAP_SHARED, hAdmxrc3, pageStart);
  if (MAP_FAILED == p) {
    switch (errno) {
      default:
        fprintf(stderr, "admxrc3MapRegion: mmap() failed, errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
    }
  }
  p += pageOffset;
  *ppVirtualBase = (void*)p;
  return ADMXRC3_SUCCESS;
}

ADMXRC3_STATUS
admxrc3UnmapRegion(
  ADMXRC3_HANDLE hAdmxrc3,
  const void* pVirtualBase)
{
  DfLinuxIoctlFindMmapping io;
  int err, ioCode;

  io.in.bByIndex = FALSE;
  io.in.variant.virtualBase = (uint64_t)(uintptr_t)pVirtualBase;
  ioCode = _IOC(_IOC_READ | _IOC_WRITE, 150, DFLINUX_IOCTLCODE_FINDMMAPPING, 0);
  err = ioctl(hAdmxrc3, ioCode, &io);
  if (0 != err) {
    if (err < 0) {
      switch (errno) {
      case EBADF:
        return ADMXRC3_INVALID_HANDLE;

      default:
        fprintf(stderr, "admxrc3UnmapRegion: ioctl() failed, errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    } else {
      switch ((DfLinuxMmapStatus)err) {
      case DfLinuxMmapInvalidRegion:
        return ADMXRC3_INVALID_REGION;

      case DfLinuxMmapInvalidTag:
        return ADMXRC3_INVALID_INDEX;

      case DfLinuxMmapNoMemory:
        return ADMXRC3_NO_MEMORY;

      case DfLinuxMmapRegionTooLarge:
        return ADMXRC3_REGION_TOO_LARGE;

      case DfLinuxMmapAccessDenied:
        return ADMXRC3_ACCESS_DENIED;

      case DfLinuxMmapGeneralFailure:
      default:
        return ADMXRC3_UNEXPECTED_ERROR;
      }
    }
  }
  if (0 == munmap((void*)(uintptr_t)io.out.virtualBase, (size_t)io.out.virtualLength)) {
    return ADMXRC3_SUCCESS;
  } else {
      switch (errno) {
      case EINVAL:
        return ADMXRC3_INVALID_REGION;

      default:
        fprintf(stderr, "admxrc3UnmapRegion: munmap() failed, errno=%d (%s)\n", errno, strerror(errno));
        return ADMXRC3_UNEXPECTED_ERROR;
      }
  }
}
