/*
** File: device.h  
** Project: ADB3 core driver
** Purpose: Data structures and definitions private to ADB3 core driver,
**          including the device context structure.
**
** (C) Copyright Alpha Data 2009-2010
*/

#if !defined(ADATA_CORE_DEVICE_H)
#define ADATA_CORE_DEVICE_H

#include <df.h>

#include "adb3.h"
#include "admxrc2.h"
#include "adpexrc5t.h"
#include "admxrc6tl.h"
#include "admxrc6t1.h"
#include "admxrc6tx.h"
#include "cfi.h"
#include "pci9xxx.h"
#include "shared_defs.h"
#include "device_defs.h"
#include "coreif.h"

struct _Adb3CoreDeviceContext;
typedef struct _Adb3CoreDeviceContext Adb3CoreDeviceContext;
struct _Adb3CoreNotificationList;
typedef struct _Adb3CoreNotificationList Adb3CoreNotificationList;
struct _ClockGeneratorContext;
typedef struct _ClockGeneratorContext ClockGeneratorContext;
struct _DmaChannelContext;
typedef struct _DmaChannelContext DmaChannelContext;
struct _WindowInfo;
typedef struct _WindowInfo WindowInfo;

typedef struct _ResourceQueue {
  CoreResourceDesc* pHead;
  CoreResourceDesc* pTail;
  unsigned int      length;
  unsigned int      users;
} ResourceQueue;

typedef enum _ClockProgramStatus {
  ClockProgramSuccess = 0,
  ClockProgramInvalidIndex = 1,
  ClockProgramHardwareError = 3
} ClockProgramStatus;

typedef enum _ClockWordStatus {
  ClockWordSuccess = 0,
  ClockWordInvalidIndex = 1,
  ClockWordOutOfRange = 2,
  ClockWordGeneralFailure = 4
} ClockWordStatus;

typedef enum _FpgaControlStatus {
  FpgaControlSuccess = 0,
  FpgaControlInvalidIndex = 1,
  FpgaControlInvalidOp = 2
} FpgaControlStatus;

typedef enum _FpgaSelectMapStatus {
  FpgaSelectMapSuccess = 0,
  FpgaSelectMapInvalidIndex = 1
} FpgaSelectMapStatus;

struct _ClockGeneratorContext {
  unsigned int index;
  uint64_t currentFrequency; /* 0 means 'not valid / not programmed / last program failed' */
  struct {
    unsigned int state;
    uint32_t word;
    DfTimer timer;
    CoreClockProgCallback* pCallback;
    CoreTicket* pCallbackTicket;
    void* pCallbackContext;
  } program;
};

struct _DmaChannelContext {
  DfSpinLock lock;
  struct {
    DfDmaBuffer* pDmaBuffer;
    union {
      void* pGeneric;
      Adb3DmaDescriptor* pAdb3;
    } buffer;
    uint64_t busAddress;
    size_t length; /* in bytes */
  } descriptor[2];
  DfDpc completionDpc;
  struct {
    CoreTicket* pTicket;
    CoreDmaStatus status;
    CoreDmaCallback* pCallback;
    void* pCallbackContext;
    CoreDmaParameters parameters;
    DfDmaMapperSlice slice;
    DfDmaMapperPosition position;
    struct {
      void* pElements;
      unsigned int length;
    } table[2];
    struct {
      boolean_t bCancel;
      unsigned int main;
      unsigned int setup;
      unsigned int setupIndex;
      unsigned int setupCount;
      unsigned int transfer;
      unsigned int transferIndex;
      unsigned int transferCount;
      unsigned int teardown;
      unsigned int teardownIndex;
      unsigned int teardownCount;
    } state;
  } curr;
};

typedef enum _DmaCompletionCode {
  DmaCompletionNormal = 0,
  DmaCompletionFailed = 1,
  DmaCompletionEot = 2
} DmaCompletionCode;

struct _WindowInfo {
  uint64_t busSize;
  uint64_t busBase;
  uint64_t localSize;
  uint64_t localBase;
  uint64_t translatedSize;
  uint64_t translatedBase;
  void* pKernelBase;
  size_t kernelSize;
};

struct _Adb3CoreNotificationList {
  DfSpinLock lock;
  DfList header;
};

struct _Adb3CoreDeviceContext {
	DfDeviceObject* pDevObj;
	boolean_t bCoreInterfaceRegd;
  DfBusResources rawResources;
  DfBusResources translatedResources;
  boolean_t bDmaMapperAllocated;

	/* This structure contains method pointers for performing model-specific actions */
	struct {
		/* The following method is responsible for initializing all of the other method pointers */
		void (*pModelBootstrap)(Adb3CoreDeviceContext* pDevCtx);

		/* These following method pointers are initialized by the ModelBootstrap method */
		void (*pEnableInterrupts)(Adb3CoreDeviceContext* pDevCtx, boolean_t bDisable);
    ClockProgramStatus (*pClockProgram)(Adb3CoreDeviceContext* pDevCtx, unsigned int index, uint32_t clockWord, boolean_t bAfter);
    ClockWordStatus (*pClockWord)(Adb3CoreDeviceContext* pDevCtx, unsigned int index, uint32_t flags, uint64_t frequencyReq, uint64_t frequencyMin, uint64_t frequencyMax, uint32_t* clockWord, uint64_t* pActualFrequency);
    void (*pDmaList)(Adb3CoreDeviceContext* pDevCtx, boolean_t bWriteToDevice, boolean_t bFixedLocal, DfDmaMapperNode* pTable, unsigned int tableLength, void* pDescriptorBuffer, uint64_t descriptorBusAddress);
    void (*pDmaTransfer)(Adb3CoreDeviceContext* pDevCtx, unsigned channel, boolean_t bStart, void* pDescriptorBuffer, unsigned int numDescriptor, uint64_t descriptorBusAddress);
		void (*pFlash)(Adb3CoreDeviceContext* pDevCtx, unsigned int bankIndex, boolean_t bWrite, uint64_t address, uint32_t* pItem);
		void (*pFlashReset)(Adb3CoreDeviceContext* pDevCtx, unsigned int bankIndex);
		FpgaControlStatus (*pFpgaControl)(Adb3CoreDeviceContext* pDevCtx, unsigned int index, CoreFpgaControlOp opCode, boolean_t* pbValue);
		FpgaSelectMapStatus (*pFpgaSelectMap)(Adb3CoreDeviceContext* pDevCtx, unsigned int index, boolean_t bWrite, size_t length, void* pBuffer);
    void (*pHwMonPump)(Adb3CoreDeviceContext* pDevCtx);
		boolean_t (*pInitHardware)(Adb3CoreDeviceContext* pDevCtx, boolean_t bInit);
    boolean_t (*pIsr)(DfInterruptObject* pInterruptObj, void* pContext);
		boolean_t (*pMapBusResources)(Adb3CoreDeviceContext* pDevCtx, DfBusResources* pRaw, DfBusResources* pTranslated, boolean_t bUnmap);
		CoreSensorStatus (*pReadSensor)(Adb3CoreDeviceContext* pDevCtx, unsigned int sensorIndex, CoreSensorValue* pReading);
		CoreVpdStatus (*pReadWriteVpd)(Adb3CoreDeviceContext* pDevCtx, boolean_t bWrite, size_t address, size_t length, void* pData);
		boolean_t (*pValidateBusResources)(Adb3CoreDeviceContext* pDevCtx, DfBusResources* pRaw, DfBusResources* pTranslated);
	} methods;

	struct {
    boolean_t bInitialized;
    boolean_t bMapped;

		union {
			void* pGeneric;
      Adb3GenericRegs* pAdb3;
      Pci9080Registers* pPci9080;
      Pci9656Registers* pPci9656;
		} bridge;
    DfMemoryHandle hBridge;

		struct {
      union {
        Adb3GenericRegs adb3;
        struct {
          uint32_t intcsrMask;
          uint32_t intcsr;
          uint16_t dmacsr;
        } pci9656;
      } bridge;
      union {
        struct {
          uint8_t fcon;
          uint8_t ienab;
        } admxrc2;
        ModelRegsAdpexrc5t adpexrc5t;
        ModelRegsAdmxrc6tx admxrc6tx; /* ADM-XRC-6TL / 6T1 */
      } model;
		} shadow;

		union {
			void* pGeneric;
      ModelRegsAdmxrc2* pAdmxrc2;
      ModelRegsAdpexrc5t* pAdpexrc5t;
      ModelRegsAdmxrc6tx* pAdmxrc6tx; /* ADM-XRC-6TL / 6T1 */
		} model;
    DfMemoryHandle hModel;

    void* pBridged;
    DfMemoryHandle hBridged;
    void* pBridgedPrefetch;
    DfMemoryHandle hBridgedPrefetch;
  } hardware;

  struct {
    DfSpinLock lock;
    struct {
      ResourceQueue queue;
      unsigned int users[MAX_NUM_CLOCK_GENERATOR];
#if defined(DF_DBG_BUILD)
      CoreTicket* pCurrentTicket[MAX_NUM_CLOCK_GENERATOR];
#endif
    } clockGen;
    struct {
      ResourceQueue queue;
      unsigned int users[MAX_NUM_DMA_ENGINE];
#if defined(DF_DBG_BUILD)
      CoreTicket* pCurrentTicket[MAX_NUM_DMA_ENGINE];
#endif
    } dmaEngine;
    struct {
      ResourceQueue queue;
      unsigned int users[1];
#if defined(DF_DBG_BUILD)
      CoreTicket* pCurrentTicket[1];
#endif
    } eeprom;
    struct {
      ResourceQueue queue;
      unsigned int users[MAX_NUM_FLASH_BANK];
#if defined(DF_DBG_BUILD)
      CoreTicket* pCurrentTicket[MAX_NUM_FLASH_BANK];
#endif
    } flash;
    struct {
      ResourceQueue queue;
      unsigned int users[MAX_NUM_TARGET_FPGA];
#if defined(DF_DBG_BUILD)
      CoreTicket* pCurrentTicket[MAX_NUM_TARGET_FPGA];
#endif
    } targetFpga;
    struct {
      ResourceQueue queue;
      unsigned int users[1];
#if defined(DF_DBG_BUILD)
      CoreTicket* pCurrentTicket[1];
#endif
    } sysMon;
  } resource;

  struct {
    struct { /* The following information must be valid when the ModelBootstrap method returns */
      unsigned int numDmaChannel;
      unsigned int numClockGen;
      unsigned int numIoModule;
      unsigned int numSensor;
      boolean_t bIncludeDiagSensors;
      unsigned int numTargetFpga;
      unsigned int numWindow;
      boolean_t bVpdInFlash;
      uint32_t vpdLength;
      uint32_t vpdOffset;
      unsigned int flashDeviceWidth; /* Width of Flash device data bus, in bytes */ 
      unsigned int flashWidth;       /* Width of Flash data actually used on the board */ 
      boolean_t bFlashStackedDie;    /* TRUE if flash device is stacked-die type */
      boolean_t b64BitDmaAddress;
      unsigned int dmaDescriptorSize;
      size_t dmaTransferMaxSize;
      boolean_t bWorkAround4kCrossing;
    } bootstrap;

    /* The following are initialized by identifyFlash */
    struct {
      boolean_t bPresent;
      CfiInformation* pDetail;
      uint64_t totalSize; /* total size (of all dies) */
      struct { /* defines start and size of target FPGA bitstream area */
        uint64_t start;
        uint64_t length;
      } targetArea;
    } flash;

    /* The following are initialized by initHardware */
    CoreModelType model;
    struct {
      union {
        uint8_t generic;
        union {
          VpdAdmxrc2Rev0 rev0;
        } admxrc2;
        union {
          VpdAdpexrc5tRev0 rev0;
        } adpexrc5t;
        union {
          VpdAdmxrc6tlRev1 rev1;
        } admxrc6tl;
        union {
          VpdAdmxrc6t1Rev0 rev0;
        } admxrc6t1;
      } data;
      boolean_t bValid;
      boolean_t bChecksumError;
      unsigned int length;
    } vpd;
    WindowInfo window[MAX_NUM_LOCAL_BUS_WINDOW];
    union {
      struct {
        uint32_t lclk;
        uint32_t mclk0, mclk1;
      } adpexrc5t;
      struct {
        uint32_t lclk;
      } admxrc6tl;
      struct {
        uint32_t lclk;
      } admxrc6t1;
    } referenceClock;
    CoreSensorInfo sensor[MAX_NUM_SENSOR];
    CoreIoModuleInfo ioModule[MAX_NUM_IO_MODULE];
  } info;

  struct {
    uint8_t* pMergeBuffer; /* buffer for merging writes into sectors */
  } flashControl;

  ClockGeneratorContext clockGenerator[MAX_NUM_CLOCK_GENERATOR];

  struct {
    Adb3CoreNotificationList fpgaInterrupt[MAX_NUM_TARGET_FPGA];
    Adb3CoreNotificationList fpgaAlert[MAX_NUM_TARGET_FPGA];
    Adb3CoreNotificationList sensorPoll;
  } notification;

  struct {
    boolean_t bConnected;
    boolean_t bEnabled;
    DfInterruptObject object;
    struct {
      DfDpc dpc;
    } fpga[MAX_NUM_TARGET_FPGA];
    struct {
      DfDpc dpc;
    } hwMon[MAX_NUM_TARGET_FPGA];
  } interrupt;

  struct {
    unsigned int mapLength;
    DmaChannelContext channel[MAX_NUM_DMA_ENGINE];
  } dma;

  struct {
    DfSpinLock lock;
    struct {
      unsigned int enable;
      DfTimer timer;
      boolean_t bTimerRunning;
      boolean_t bShutdown;
      DfEvent shutdownEvent;
    } polling;
    struct {
      CoreSensorValue value[MAX_NUM_SENSOR];
      uint8_t enable[MAX_NUM_SENSOR];
    } history;
    union {
      struct {
        unsigned int state;
      } admxrc6tl;
      struct {
        unsigned int state;
      } admxrc6t1;
    } model;
  } hwMon;
};

/* Returns true if the range of items specified by [ START , START + N ) is not a subset of [ 0 , LENGTH ) */
#define CHECK_BOUNDS(START, N, LENGTH) \
	((START) + (N) < (START) || (START) + (N) > (LENGTH))

/* Returns true if the range of items specified by [ START , START + N ) is not a subset of [ OFFSET , OFFSET + LENGTH ) */
#define CHECK_BOUNDS_OFFSET(START, N, OFFSET, LENGTH) \
	((START) + (N) < (START) || (START) < (OFFSET) || (START) + (N) > (LENGTH) + (OFFSET))

#endif
