/*
** File: model_adpexrc5t.c  
** Project: ADB3 core driver
** Purpose: ADPe-XRC-5T-specific code, including function to bootstrap device context.
**
** (C) Copyright Alpha Data 2009-2010
**
** NOTES
**
** 1. Organization of Flash device U68 (all offsets in bytes):
**
**   0x000000 - 0x4BFFFF  Bridge bitstream (sufficient for 5VFX100T, 5VLX110T, 5VSX95T)
**   0x4C0000 - 0x4FFFFF  VPD
**   0x500000 - 0x500003  Target bitstream boot flag
**   0x500004 - 0xFFFFFF  Target bitstream (sufficient for 5VFX200T, 5VLX330T, 5VSX240T)
*/

#include <df.h>

#include "adb3.h"
#include "adb3_common.h"
#include "adpexrc5t.h"
#include "coredma.h"
#include "ics843034_01.h"
#include "device.h"
#include "model_boot.h"
#include "flash.h"
#include "flash_cfi.h"

static void
createFakeVpd(
	Adb3CoreDeviceContext* pDevCtx)
{
  VpdAdpexrc5tRev0* pVpd = &pDevCtx->info.vpd.data.adpexrc5t.rev0;
  int i;

  DF_ZERO_OBJECT(pVpd);
  pVpd->version = 0;
  pVpd->length = sizeof(*pVpd);
  pVpd->serial = 1;
  pVpd->modifications = 0;
  pVpd->lclkRef = 200000000;
  pVpd->mclkRef0 = 25000000;
  pVpd->mclkRef1 = 26562500;
  for (i = 0; i < 4; i++) {
    pVpd->sdram[i].order = 26;
    pVpd->sdram[i].speed = 3333;
    pVpd->sdram[i].link = 0;
  }
  for (; i < 6; i++) {
    pVpd->ssram[i - 4].order = 20;
    pVpd->ssram[i - 4].speedMax = 2000;
    pVpd->ssram[i - 4].speedMin = 1200;
    pVpd->ssram[i - 4].link = 0;
  }
  pVpd->fpga.device = 0x8BU; /* 5VLX110T */
  dfAssert(sizeof(char) == sizeof(uint8_t));
  dfStrCpy((char*)pVpd->fpga.speedGrade, DF_ARRAY_LENGTH(pVpd->fpga.speedGrade), "1");
  dfStrCpy((char*)pVpd->fpga.stepping, DF_ARRAY_LENGTH(pVpd->fpga.stepping), "ES");
  pVpd->fpga.tempGrade = ADMXRC6TL_VPD_TEMPGRADE_COMMERCIAL;
  pVpd->pcbRev = 0x10U;
  pVpd->checksum = 0U;

  pDevCtx->info.vpd.bValid = TRUE;
  pDevCtx->info.vpd.bChecksumError = TRUE;
  pDevCtx->info.vpd.length = sizeof(pDevCtx->info.vpd.data.adpexrc5t.rev0);
}

static ClockProgramStatus
clockProgram(
	Adb3CoreDeviceContext*	pDevCtx,
  unsigned int clockIndex,
	uint32_t clockWord,
	boolean_t bAfter)
{
	ModelRegsAdpexrc5t* pModelRegs;
  DfMemoryHandle hModel;
	uint32_t val32, mask;

	dfDebugPrint(5, ("clockProgram: entered, clockIndex=%lu clockWord=0x%lx bAfter=%s\n",
    (unsigned long)clockIndex, (unsigned long)clockWord, bAfter ? "TRUE" : "FALSE"));
	
  pModelRegs = pDevCtx->hardware.model.pAdpexrc5t;
  hModel = pDevCtx->hardware.hModel;

	if (bAfter) {
    switch (clockIndex) {
    case 0:
			val32 = dfPciMemRead32(hModel, &pModelRegs->lclkStat3);
  	  dfDebugPrint(5, ("clockProgram: LCLKSTAT3=0x%lx\n", (unsigned long)val32));	
			mask = (val32 & 0x4U) ? 0x2U : 0x1U;
			if (!(val32 & mask)) {
				/* DCM not locked */
				return ClockProgramHardwareError;
			}
			if (val32 & 0x28U) {
				/* Control logic busy or write of DCM parameters timed out */
				return ClockProgramHardwareError;
			}
			if (val32 & 0x10U) {
				/* Error */
				return ClockProgramHardwareError;
			}
      break;

    case 1:
			val32 = dfPciMemRead32(hModel, &pModelRegs->mclkCtl);
  	  dfDebugPrint(5, ("clockProgram: mclkCtl=0x%lx\n", (unsigned long)val32));	
			if (!(val32 & 0x80000000U)) {
				return ClockProgramHardwareError;
			}
      break;

    default:
      return ClockProgramInvalidIndex;
		}
	} else {
    switch (clockIndex) {
    case 0:
			dfPciMemWrite32(hModel, &pModelRegs->lclkCtl, clockWord | (0x1U << 31));
      dfPciMemRead32(hModel, &pModelRegs->lclkStat3);
      break;

    case 1:
      dfPciMemWrite32(hModel, &pModelRegs->mclkCtl, (0x1U << 31) | ((clockWord & 0x20000U) << 13) | (clockWord & 0x1ffffU));
      dfPciMemRead32(hModel, &pModelRegs->mclkCtl);
      break;

    default:
      return ClockProgramInvalidIndex;
    }
	}

	return ClockProgramSuccess;
}

static ClockWordStatus
clockWordLclk(
	Adb3CoreDeviceContext* pDevCtx,
  uint32_t flags,
	uint32_t refFrequency,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
	uint32_t* pClockWord,
	uint64_t* pActualFrequency)
{
	const uint32_t minFrequency = 31999000, maxFrequency = 210001000;
	uint32_t m, bestM = 0, d, bestD = 0; /* init. to avoid compiler whine */
	uint32_t f, bestFrequency = 0; /* init. to avoid compiler whine */
	uint32_t error, bestError = (uint32_t)-1;

	dfDebugPrint(5, ("clockWordLclk: refFrequency=%lu reqFrequency=%llu\n",
    (unsigned long)refFrequency, (unsigned long long)frequencyReq));

  if (refFrequency == 0 || refFrequency > 268000000) {
		return ClockWordGeneralFailure;
	}
	if (frequencyReq < minFrequency || frequencyReq > maxFrequency) {
		return ClockWordOutOfRange;
	}

	/* Won't overflow assuming (refFrequency <= 268e6) */
	for (d = 1; d <= 32; d++) {
		for (m = 2; m <= 32; m++) {
			f = ((refFrequency >> 1) * m / d) << 1;
      if ((flags & CoreClockMinimumFrequency) && f < frequencyMin) {
        continue;
      }
      if ((flags & CoreClockMaximumFrequency) && f > frequencyMax) {
        continue;
      }
			if (f > (uint32_t)frequencyReq) {
				error = f - (uint32_t)frequencyReq;
			} else {
				error = (uint32_t)frequencyReq - f;
			}
			if (error < bestError) {
				bestError = error;
				bestFrequency = f;
				bestM = m;
				bestD = d;
			}
		}
	}

	if (bestError == (uint32_t)-1) {
		/* Failed to find a suitable value */
		return ClockWordGeneralFailure;
	}

	*pActualFrequency = bestFrequency;
	*pClockWord = ((bestD - 1) << 16) | (bestM - 1);

	dfDebugPrint(5, ("clockWordLclk: bestFrequency=%lu clockWord=0x%08lx\n", (unsigned long)bestFrequency, (unsigned long)*pClockWord));

	return ClockWordSuccess;
}

static ClockWordStatus
clockWord(
	Adb3CoreDeviceContext* pDevCtx,
	unsigned int clockIndex,
  uint32_t flags,
	uint64_t frequencyReq,
	uint64_t frequencyMin,
	uint64_t frequencyMax,
	uint32_t* pClockWord,
	uint64_t* pActualFrequency)
{
  switch (clockIndex) {
  case 0:
    return clockWordLclk(pDevCtx, flags, pDevCtx->info.referenceClock.adpexrc5t.lclk / 2, frequencyReq, frequencyMin, frequencyMax, pClockWord, pActualFrequency);

  case 1:
    return clockWordIcs843034_01(pDevCtx, flags, pDevCtx->info.referenceClock.adpexrc5t.mclk0, pDevCtx->info.referenceClock.adpexrc5t.mclk1, frequencyReq, frequencyMin, frequencyMax, pClockWord, pActualFrequency);

  default:
    return ClockWordInvalidIndex;
	}
}

static void
enableInterrupts(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bEnable)
{
	Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;

  dfDebugPrint(1, ("enableInterrupts: entered bEnable=%s\n", bEnable ? "TRUE" : "FALSE"));

	if (bEnable) {
    pDevCtx->hardware.shadow.bridge.adb3.irqEnable = ADB3IRQ_MODEL | ADB3IRQ_DMA_ALL;
    dfPciMemWrite32(hBridge, &pRegs->irqEnable, pDevCtx->hardware.shadow.bridge.adb3.irqEnable);
		dfDebugPrint(2, ("enableInterrupts: interrupts enabled, irqEnable=0x%08lx\n", (unsigned long)dfPciMemRead32(hBridge, &pRegs->irqEnable)));
  } else {
    pDevCtx->hardware.shadow.bridge.adb3.irqEnable = 0;
    dfPciMemWrite32(hBridge, &pRegs->irqEnable, pDevCtx->hardware.shadow.bridge.adb3.irqEnable);
		dfDebugPrint(2, ("enableInterrupts: interrupts disabled, irqEnable=0x%08lx\n", (unsigned long)dfPciMemRead32(hBridge, &pRegs->irqEnable)));
	}
}

static FpgaControlStatus
fpgaControl(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  CoreFpgaControlOp opCode,
  boolean_t* pbValue)
{
  ModelRegsAdpexrc5t* pModelRegs;
  DfMemoryHandle hModel;
  uint32_t val32;
  boolean_t bOk;

  dfDebugPrint(8, ("fpgaControl: index=%lu opCode=%lu pbValue=0x%p\n",
    (unsigned long)index, (unsigned long)opCode, pbValue));
  
  if (0 != index) {
    return FpgaControlInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdpexrc5t;
  hModel = pDevCtx->hardware.hModel;

  switch (opCode) {
  case CoreFpgaControlOpGetDone:
    bOk = TRUE;
    val32 = dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);
    if (!(val32 & 0x4U)) {
      dfDebugPrint(9, ("fpgaControl: fpgaStat=0x%lx\n", (unsigned long)val32));
      bOk = FALSE;
    }
    *pbValue = bOk;
    dfDebugPrint(9, ("fpgaControl: opCode=GetDone val32=0x%lx *pbValue=%s\n", (unsigned long)val32, bOk ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpGetLinkStatus:
    bOk = TRUE;
    /* Check that MPTL is up */
    val32 = dfPciMemRead32(hModel, &pModelRegs->mptlStat);
    if ((val32 & 0x7U) != 0x7U) {
      dfDebugPrint(9, ("fpgaControl: mptlStat=0x%lx\n", (unsigned long)val32));
      bOk = FALSE;
    }
    *pbValue = bOk;
    dfDebugPrint(9, ("fpgaControl: opCode=GetLinkStat val32=0x%lx *pbValue=%s\n", (unsigned long)val32, bOk ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpGetInit:
    val32 = dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);
    *pbValue = (val32 & 0x2U) ? TRUE : FALSE;
    dfDebugPrint(9, ("fpgaControl: opCode=GetInit val32=0x%lx *pbValue=%s\n", (unsigned long)val32, *pbValue ? "TRUE" : "FALSE"));
    break;

  case CoreFpgaControlOpSetProg:
    val32 = pDevCtx->hardware.shadow.model.adpexrc5t.fpgaCtl;
    if (*pbValue) {
      val32 |= 0x1U;
    } else {
      val32 &= ~0x1U;
    }
    dfDebugPrint(9, ("fpgaControl: opCode=SetProg val32=0x%lx *pbValue=%s\n", (unsigned long)val32, *pbValue ? "TRUE" : "FALSE"));
    dfPciMemWrite32(hModel, &pModelRegs->fpgaCtl, val32);
    dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);
    break;

  default:
    return FpgaControlInvalidOp;
  }

  return FpgaControlSuccess;
}

static FpgaSelectMapStatus
fpgaSelectMap(
  Adb3CoreDeviceContext* pDevCtx,
  unsigned int index,
  boolean_t bWrite,
  size_t length,
  void* pBuffer)
{
  const size_t mask32 = 0x3U;
  ModelRegsAdpexrc5t* pModelRegs;
  DfMemoryHandle hModel;
  uint8_t* pBuffer8;
  uint32_t* pBuffer32;
  size_t length32, align32;
  uint8_t val8;
  uint32_t val32;

  dfDebugPrint(8, ("fpgaSelectMap: index=%lu bWrite=%s length=%llu(0x%08lx_%08lx) pBuffer=0x%p\n",
    (unsigned long)index, bWrite ? "TRUE" : "FALSE", (unsigned long long)length, dfSplitUint64((uint64_t)length), pBuffer));
  
  if (0 != index) {
    return FpgaSelectMapInvalidIndex;
  }

  pModelRegs = pDevCtx->hardware.model.pAdpexrc5t;
  hModel = pDevCtx->hardware.hModel;
  pBuffer8 = (uint8_t*)pBuffer;

	if (bWrite) {
    /* Align pointer to 32-bit boundary using 8-bit transfers */
    align32 = (size_t)(uintptr_t)pBuffer & mask32;
    length -= align32;
    while (align32--) {
			val8 = *pBuffer8++;
			dfPciMemWrite8(hModel, &pModelRegs->selectMap[0], val8);
    }

    /* Do bulk of data using 32-bit transfers */
    length32 = length & ~mask32;
    pBuffer32 = (uint32_t*)pBuffer8;
    length -= length32;
    while (length32) {
			val32 = *pBuffer32++;
			dfPciMemWrite32(hModel, (uint32_t*)&pModelRegs->selectMap[0], val32);
      length32 -= 4;
    }

    /* Do remainder using 8-bit transfers */
    pBuffer8 = (uint8_t*)pBuffer32;
		while (length--) {
			val8 = *pBuffer8++;
			dfPciMemWrite8(hModel, &pModelRegs->selectMap[0], val8);
		}

		dfPciMemRead32(hModel, &pModelRegs->fpgaCtl); /* Ensure above writes have propagated to destination */
	} else {
		while (length--) {
			val8 = dfPciMemRead8(hModel, &pModelRegs->selectMap[0]);
			*pBuffer8++ = val8;
		}
	}

  return FpgaSelectMapSuccess;
}

static boolean_t
initHardware(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bInit)
{
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  ModelRegsAdpexrc5t* pModelRegs;
  DfMemoryHandle hModel;
  unsigned int i;

  dfDebugPrint(2, ("initHardware: bInit=%s\n", bInit ? "TRUE" : "FALSE"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;
  pModelRegs = pDevCtx->hardware.model.pAdpexrc5t;
  hModel = pDevCtx->hardware.hModel;

  if (bInit) {
    pDevCtx->info.model = CoreModelAdpexrc5t;

    /* Initialize DMA enables */
    dfPciMemWrite32(hBridge, &pRegs->dmaEnable, ADB3DMAEN_LOCAL_MASTER | ADB3DMAEN_DMA_ALL);

    /* Allow up to 3 PCIe packets in flight at a time */
    dfPciMemWrite32(hBridge, &pRegs->pcieControl, 0x02100100U);           

    /* Initialize DMA engine interrupt enable registers */
    for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
      /* Request up to 32 x 16-byte words on PCIe side */
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].hostControl, 0x20U);    
      /* Allow up to 64 bytes read and 64 bytes written on MPTL */
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[i].fpgaControl, 0x80008U);
    }

    /* Initialize FPGACTL register for 8-bit SelectMap port */
    pDevCtx->hardware.shadow.model.adpexrc5t.fpgaCtl = 0;
    dfPciMemWrite32(hModel, &pModelRegs->fpgaCtl, pDevCtx->hardware.shadow.model.adpexrc5t.fpgaCtl);
		dfPciMemRead32(hModel, &pModelRegs->fpgaCtl);

    /* Identify Flash device */
    if (pDevCtx->info.bootstrap.flashWidth != 0) {
      if (flashCfiIdentify(pDevCtx, 0, 0)) {
        pDevCtx->info.flash.bPresent = TRUE;
        pDevCtx->info.flash.targetArea.start = 0x500000U;
        pDevCtx->info.flash.targetArea.length = pDevCtx->info.flash.totalSize - pDevCtx->info.flash.targetArea.start;
      } else {
        dfDebugPrint(0, ("*** initHardware: failed to identify Flash\n"));
      }
    }

    createFakeVpd(pDevCtx);

    /* Divide LCLK reference clock by 2 because on the ADPE-XRC-5T, a CLKDIV value of 2 is used in the DCMs that generate LCLK */
    pDevCtx->info.referenceClock.adpexrc5t.lclk = pDevCtx->info.vpd.data.adpexrc5t.rev0.lclkRef;
    pDevCtx->info.referenceClock.adpexrc5t.mclk0 = pDevCtx->info.vpd.data.adpexrc5t.rev0.mclkRef0;
    pDevCtx->info.referenceClock.adpexrc5t.mclk1 = pDevCtx->info.vpd.data.adpexrc5t.rev0.mclkRef1;

    /* Find base addresses and sizes of local bus windows */
    pDevCtx->info.window[0].busBase = pDevCtx->rawResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].busSize = pDevCtx->rawResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].localBase = 0;
    pDevCtx->info.window[0].localSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[2].base;
    pDevCtx->info.window[0].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[0].pKernelBase = pDevCtx->hardware.pBridged;
    pDevCtx->info.window[0].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[2].size;
    pDevCtx->info.window[1].busBase = pDevCtx->rawResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].busSize = pDevCtx->rawResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].localBase = 0;
    pDevCtx->info.window[1].localSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[3].base;
    pDevCtx->info.window[1].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[1].pKernelBase = pDevCtx->hardware.pBridgedPrefetch;
    pDevCtx->info.window[1].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[3].size;
    pDevCtx->info.window[2].busBase = pDevCtx->rawResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[2].busSize = pDevCtx->rawResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[2].localBase = 0;
    pDevCtx->info.window[2].localSize = 0;
    pDevCtx->info.window[2].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[1].base;
    pDevCtx->info.window[2].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[2].pKernelBase = pDevCtx->hardware.model.pGeneric;
    pDevCtx->info.window[2].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[1].size;
    pDevCtx->info.window[3].busBase = pDevCtx->rawResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[3].busSize = pDevCtx->rawResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[3].localBase = 0;
    pDevCtx->info.window[3].localSize = 0;
    pDevCtx->info.window[3].translatedBase = pDevCtx->translatedResources.variant.pci.memBar[0].base;
    pDevCtx->info.window[3].translatedSize = pDevCtx->translatedResources.variant.pci.memBar[0].size;
    pDevCtx->info.window[3].pKernelBase = pDevCtx->hardware.bridge.pGeneric;
    pDevCtx->info.window[3].kernelSize = (size_t)pDevCtx->translatedResources.variant.pci.memBar[0].size;
  } else {
#if 0
    /* Initialize DMA engine interrupt enable registers */
    for (i = 0; i < pDevCtx->info.bootstrap.numDmaChannel; i++) {
      dfPciMemWrite32(&pRegs->dmaEngine[i].irqEnable, 0);
    }
#endif
    
    /* Deinitialize DMA enables */
    dfPciMemWrite32(hBridge, &pRegs->dmaEnable, 0);
  }

  return TRUE;
}

static boolean_t
isr(
  DfInterruptObject* pInterruptObj,
  void* pInterruptContext)
{
  Adb3CoreDeviceContext* pDevCtx = (Adb3CoreDeviceContext*)pInterruptContext;
  Adb3GenericRegs* pRegs;
  DfMemoryHandle hBridge;
  uint32_t irqStatus, dmaIrqStatus;
  unsigned int channel, n;

  dfDebugPrint(8, ("isr: entered\n"));

  pRegs = pDevCtx->hardware.bridge.pAdb3;
  hBridge = pDevCtx->hardware.hBridge;

  /* Sample the currently active interrupts, and mask the value with the interrupts we want */
  irqStatus = dfPciMemRead32(hBridge, &pRegs->irqStatus) & pDevCtx->hardware.shadow.bridge.adb3.irqEnable;

  /* Clear the interrupts we saw */
  dfPciMemWrite32(hBridge, &pRegs->irqStatus, irqStatus);

  if (irqStatus & ADB3IRQ_MODEL) {
    dfDebugPrint(9, ("isr: MODEL interrupt\n"));
    /* TO DO */
  }

  n = pDevCtx->info.bootstrap.numDmaChannel;
  for (channel = 0; channel < n; channel++) {
    if (irqStatus & (ADB3IRQ_DMA0 << channel)) {
      dmaIrqStatus = dfPciMemRead32(hBridge, &pRegs->dmaEngine[channel].irq.status);
      dfDebugPrint(9, ("isr: DMA%lu interrupt, DMA%lu.irqStatus=0x%08lx\n",
        (unsigned long)channel, (unsigned long)channel, (unsigned long)dmaIrqStatus));
      dfPciMemWrite32(hBridge, &pRegs->dmaEngine[channel].irq.ack, dmaIrqStatus);
      dfDpcSchedule(&pDevCtx->dma.channel[channel].completionDpc, (void*)(uintptr_t)channel);
    }
  }

  return irqStatus ? TRUE : FALSE;
}

static boolean_t
mapBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated,
	boolean_t bUnmap)
{
	dfDebugPrint(1, ("mapBusResources: bUnmap=%s\n", bUnmap ? "TRUE" : "FALSE"));

	if (bUnmap) {
		if (NULL != pDevCtx->hardware.bridge.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.bridge.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 generic registers\n"));
			}
      pDevCtx->hardware.bridge.pGeneric = NULL;
		}
		if (NULL != pDevCtx->hardware.model.pGeneric) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.model.pGeneric)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 model-specific registers\n"));
			}
      pDevCtx->hardware.model.pGeneric = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridged) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridged)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 bridged space\n"));
			}
      pDevCtx->hardware.pBridged = NULL;
		}
    if (NULL != pDevCtx->hardware.pBridgedPrefetch) {
			if (!dfIoSpaceUnmap(pDevCtx->pDevObj, pDevCtx->hardware.pBridgedPrefetch)) {
				dfDebugPrint(0, ("*** mapBusResources: failed to unmap ADB3 prefetchable bridged space\n"));
			}
      pDevCtx->hardware.pBridgedPrefetch = NULL;
		}
		dfDebugPrint(2, ("mapBusResources: unmapped\n"));
	} else {
		pDevCtx->hardware.bridge.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 0, 0, pTranslated->variant.pci.memBar[0].size, &pDevCtx->hardware.hBridge);
		if (NULL == pDevCtx->hardware.bridge.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 generic registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 generic registers @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[0].base), pDevCtx->hardware.bridge.pGeneric));
		}
		pDevCtx->hardware.model.pGeneric = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 1, 0, pTranslated->variant.pci.memBar[1].size, &pDevCtx->hardware.hModel);
		if (NULL == pDevCtx->hardware.model.pGeneric) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 model-specific registers\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 model-specific registers @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[1].base), pDevCtx->hardware.model.pGeneric));
		}
    pDevCtx->hardware.pBridged = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 2, 0, pTranslated->variant.pci.memBar[2].size, &pDevCtx->hardware.hBridged);
		if (NULL == pDevCtx->hardware.pBridged) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 bridged space @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[2].base), pDevCtx->hardware.pBridged));
		}
    pDevCtx->hardware.pBridgedPrefetch = dfIoSpaceMap(pDevCtx->pDevObj, pTranslated, 3, 0, pTranslated->variant.pci.memBar[3].size, &pDevCtx->hardware.hBridgedPrefetch);
		if (NULL == pDevCtx->hardware.pBridgedPrefetch) {
			dfDebugPrint(0, ("*** mapBusResources: failed to map ADB3 prefetchable bridged space\n"));
			return FALSE;
		} else {
			dfDebugPrint(2, ("mapBusResources: mapped ADB3 prefetchable bridged space @ 0x%08lx_%08lx to 0x%p\n",
        dfSplitUint64(pTranslated->variant.pci.memBar[3].base), pDevCtx->hardware.pBridgedPrefetch));
		}
		dfDebugPrint(2, ("mapBusResources: mapped OK\n"));
	}

	return TRUE;
}

static CoreVpdStatus
readWriteVpd(
	Adb3CoreDeviceContext* pDevCtx,
	boolean_t bWrite,
	size_t address,
	size_t length,
	void* pData)
{
  CoreFlashStatus flashStatus;

  if (CHECK_BOUNDS(address, length, pDevCtx->info.flash.totalSize)) {
    return CoreVpdInvalidRegion;
  }
	if (bWrite) {
    flashStatus = flashWrite(pDevCtx, 0, address, length, pData);
	} else {
		flashStatus = flashRead(pDevCtx, 0, address, length, pData);
	}
  if (CoreFlashInvalidIndex == flashStatus) {
    /* Should never happen, but map CoreFlashInvalidIndex to CoreVpdGeneralFailure */
    return CoreVpdGeneralFailure;
  } else {
    return (CoreVpdStatus)flashStatus;
  }
}

static boolean_t
validateBusResources(
	Adb3CoreDeviceContext* pDevCtx,
	DfBusResources* pRaw,
	DfBusResources* pTranslated)
{
	dfDebugPrint(1, ("validateBusResources: entered\n"));

	if (pTranslated->variant.pci.numMemBar < 4) {
		dfDebugPrint(0, ("*** validateBusResources: less than 4 memory BARs\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[0].size < 0x1000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR0 smaller than 0x1000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[1].size < sizeof(ModelRegsAdpexrc5t)) {
		dfDebugPrint(0, ("*** validateBusResources: BAR1 smaller than 0x%lu bytes\n", (unsigned long)sizeof(ModelRegsAdpexrc5t)));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[2].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR2 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.memBar[3].size < 0x100000U) {
		dfDebugPrint(0, ("*** validateBusResources: BAR3 smaller than 0x100000 bytes\n"));
		return FALSE;
	}
	if (pTranslated->variant.pci.numIrq != 1) {
		dfDebugPrint(0, ("*** validateBusResources: not exactly 1 interrupt\n"));
		return FALSE;
	}

	dfDebugPrint(2, ("validateBusResources: OK\n"));

	return TRUE;
}

void
adb3CoreBootstrapAdpexrc5t(
	Adb3CoreDeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("adb3CoreBootstrapAdpexrc5t: entered\n"));

  pDevCtx->methods.pClockProgram = clockProgram;
  pDevCtx->methods.pClockWord = clockWord;
  pDevCtx->methods.pDmaList = adb3DmaList;
  pDevCtx->methods.pDmaTransfer = adb3DmaTransfer;
  pDevCtx->methods.pEnableInterrupts = enableInterrupts;
  pDevCtx->methods.pFpgaControl = fpgaControl;
  pDevCtx->methods.pFpgaSelectMap = fpgaSelectMap;
  pDevCtx->methods.pInitHardware = initHardware;
  pDevCtx->methods.pIsr = isr;
  pDevCtx->methods.pMapBusResources = mapBusResources;
  pDevCtx->methods.pReadWriteVpd = readWriteVpd;
  pDevCtx->methods.pValidateBusResources = validateBusResources;

  pDevCtx->info.bootstrap.numClockGen = 2;
  pDevCtx->info.bootstrap.numDmaChannel = 2;
  pDevCtx->info.bootstrap.numSensor = 0;
  pDevCtx->info.bootstrap.numTargetFpga = 1;
  pDevCtx->info.bootstrap.numWindow = 4;
  pDevCtx->info.bootstrap.bVpdInFlash = TRUE;
  pDevCtx->info.bootstrap.vpdLength = 0x40000U;
  pDevCtx->info.bootstrap.vpdOffset = 0x4c0000U;
#if 0
  pDevCtx->info.bootstrap.flashDeviceWidth = 2;
  pDevCtx->info.bootstrap.flashWidth = 2;
#else
  pDevCtx->info.bootstrap.flashDeviceWidth = 0;
  pDevCtx->info.bootstrap.flashWidth = 0;
#endif
  pDevCtx->info.bootstrap.bFlashStackedDie = FALSE;
  pDevCtx->info.bootstrap.b64BitDmaAddress = FALSE;
  dfParameterGetBoolean(pDevCtx->pDevObj->pDrvObj, "PciAddress64Bit", &pDevCtx->info.bootstrap.b64BitDmaAddress);
  pDevCtx->info.bootstrap.dmaDescriptorSize = 32;
  pDevCtx->info.bootstrap.dmaTransferMaxSize = 0x80000000U;
}
