/*
** File: userif.c  
** Project: Linux Driver Framework
** Purpose: Implements mechanism by which OS-independent driver code can register
**          an interface accessible by user-mode programs.
**
** (C) Copyright Alpha Data 2010
**
** Notes:
**
** 1. We store a pointer to a DfMmappedRegion in the VMA's vm_private_data
**    member. When the VMA's close() method is called as a result of munmap(),
**    we can remove the mapped region from the client object's list.
**
** 2. fork() causes a VMA's open() method to be called with the new VMA. There
**    is no simple way to get the old VMA. How do we ensure that, if a VMA is
**    cloned due to a fork(), and device handle has long since been closed,
**    the device whose BARs that the VMA references persists?
**
** 3. close() on a device handle does not cause any existing mmap'ed regions
**    to be munmapped.
**
** 4. mremap() can make holes in an mmap'ed region. This will not make the
**    system vulnerable to crashing.
**
** 5. mremap() can be used to extend an mmap'ed region beyond the end of a BAR
**    in a device that the VMA references. In order to prevent this resulting
**    in a  crash vulnerability, the nopage() VMA method should always return
**    VM_FAULT_SIGBUS, thus preventing invalid physical memory being accessed.
**
** 6. The fopsRelease() function is not called until the last outstanding
**    operation on that handle is completed AND any references to the
**    (struct file*) are removed.
**
**    Outstanding operations on a handle include user-space calls such as
**    ioctl(), read(), write(), mmap() etc.
**
**    A vm_area_struct that was created by an mmap() call through the
**    (struct file*) counts as a reference to the (struct file*), as does
**    duplicating a file descriptor, e.g. using dup().
**
**    The implication of this is that (unlike Windows and the IRP_MJ_CLEANUP
**    IRP), no special cancellation of client requests is required inside
**    fopsRelease(), because it won't be called until all client requests are
**    finished.
*/

#include "df.h"
#include "dflinux.h"
#include <ioctl_dflinux.h>

/* Client object state, used to wait for outstanding requests to be finished when closing */
typedef enum _ClientState {
  ClientStateOpen = 0,    /* Not being closed */
  ClientStateClosing = 1, /* Being closed, before testing if there are outstanding client requests */
  ClientStateWaiting = 2  /* Waiting for outstanding client requests to be completed */
} ClientState;

typedef enum _NonBlockState {
  NonBlockStateIdle = 0,     /* No non-blocking operations in progress */
  NonBlockStateIoctl = 1,    /* Executing IOCTL handler */
  NonBlockStateDeferred = 2  /* Waiting on completion */
} NonBlockState;

static uint MaxNumMinor = 16;
module_param(MaxNumMinor, uint, S_IRUGO);

/* Semaphore to control access to VMAs */
struct semaphore g_vmaSemaphore;

/*
** ---------------------------------------------------------------------------
** Functions for managing client request lists
** ---------------------------------------------------------------------------
*/

static inline void
requestInit(
  DfClientRequest* pReq,
  DfClientObject* pClObj)
{
  pReq->system.pClObj = pClObj;
  pReq->system.pCancelCallback = NULL;
  pReq->system.bCancel = FALSE;
  pReq->system.bCompleted = FALSE;
  pReq->system.pContext = NULL;
  pReq->system.pWaitQueue = NULL;
}

/* Uninterruptibly wait for a client request to be completed */
static void
requestWait(
  DfClientRequest* pReq)
{
  DfClientObject* pClObj = pReq->system.pClObj;
  DfSpinLockFlags f;
  wait_queue_head_t waitQueue;
  wait_queue_t wait;

  spin_lock_irqsave(&pClObj->system.lock, f);
  if (pReq->system.bCompleted) {
    spin_unlock_irqrestore(&pClObj->system.lock, f);
  } else {
    pReq->system.pWaitQueue = &waitQueue;
    init_waitqueue_head(&waitQueue);
    init_wait(&wait);
    add_wait_queue(&waitQueue, &wait);
    set_current_state(TASK_UNINTERRUPTIBLE);
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    schedule();
    set_current_state(TASK_RUNNING);
    remove_wait_queue(&waitQueue, &wait);
  }
}

static DfIoStatus
deliverRequestBlocking(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfOnIoctlMethod* pHandler,
  unsigned int reqCode,
  void* pIoctlStruct,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  struct {
    uint8_t data[DFLINUX_MAX_IOCTL_SIZE];
  } ioctlStruct;
  DfClientRequest* pReq;
  DfIoStatus status;
  wait_queue_head_t waitQueue;
  wait_queue_t wait;
  DfSpinLockFlags f;
  DfCancelCallback* pCancelCallback;

  pReq = dfPoolAlloc(&pDevObj->system.request.pool, DfClientRequest);
  if (NULL == pReq) {
    return -ENOMEM;
  }
  requestInit(pReq, pClObj);

  if (copy_from_user(&ioctlStruct, pIoctlStruct, sizeIn)) {
    dfDebugPrint(5, ("deliverRequestBlocking: fault copying ioctl struct from user space, pIoctl=0x%p sizeIn=%u\n", pIoctlStruct, sizeIn));
    status = -EFAULT;
  } else {
    status = pHandler(pDevObj, pClObj, pReq, reqCode, &ioctlStruct, sizeIn, sizeOut);
    if (status == DfIoStatusDeferred) {
      spin_lock_irqsave(&pClObj->system.lock, f);
      if (pReq->system.bCompleted) {
        spin_unlock_irqrestore(&pClObj->system.lock, f);
        status = pReq->system.status;
      } else {
        set_current_state(TASK_INTERRUPTIBLE);
        pReq->system.pWaitQueue = &waitQueue;
        init_waitqueue_head(&waitQueue);
        init_wait(&wait);
        add_wait_queue(&waitQueue, &wait);
        spin_unlock_irqrestore(&pClObj->system.lock, f);
        schedule();
        if (signal_pending(current)) {
          spin_lock_irqsave(&pClObj->system.lock, f);
          if (pReq->system.bCompleted) {
            spin_unlock_irqrestore(&pClObj->system.lock, f);
          } else {
            set_current_state(TASK_UNINTERRUPTIBLE);
            if (pReq->system.bCancel) {
              /* Somebody else is cancelling this - let them do it */
              spin_unlock_irqrestore(&pClObj->system.lock, f);
            } else {
              pReq->system.bCancel = TRUE;
              pCancelCallback = pReq->system.pCancelCallback;
              if (NULL != pCancelCallback) {
                dfDebugPrint(2, ("deliverRequestBlocking: cancelling request 0x%p\n", pReq));
                pReq->system.pCancelCallback = NULL;
                dfListRemove(&pReq->system.node);
              }
              spin_unlock_irqrestore(&pClObj->system.lock, f);
              if (NULL != pCancelCallback) {
                pCancelCallback(pReq, DfIoStatusCancelled);
              }
            }
            schedule();
          }
          status = -ECANCELED;
        } else {
          status = pReq->system.status;
        }
        remove_wait_queue(&waitQueue, &wait);
        set_current_state(TASK_RUNNING);
      }
    }
  }
  dfPoolFree(&pDevObj->system.request.pool, pReq);

  if (DfIoStatusSuccess == status && 0 != sizeOut) {
    if (copy_to_user(pIoctlStruct, &ioctlStruct, sizeOut)) {
      dfDebugPrint(5, ("deliverRequestBlocking: fault copying ioctl struct to user space, pIoctl=0x%p sizeOut=%u\n", pIoctlStruct, sizeOut));
      status = -EFAULT;
    }
  }

  return status;
}

static DfIoStatus
deliverRequestNonBlocking(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  DfOnIoctlMethod* pHandler,
  unsigned int reqCode,
  void* pIoctlStruct,
  unsigned int sizeIn,
  unsigned int sizeOut)
{
  DfIoStatus status;
  DfLinuxClientRequestNonBlock* pReq;
  DfSpinLockFlags f;

  pReq = dfPoolAlloc(&pDevObj->system.request.poolNonBlock, DfLinuxClientRequestNonBlock);
  if (NULL == pReq) {
    return -ENOMEM;
  }
  requestInit(&pReq->clientRequest, pClObj);

  if (copy_from_user(&pReq->ioctlStruct, pIoctlStruct, sizeIn)) {
    dfDebugPrint(5, ("deliverRequestNonBlocking: fault copying ioctl struct from user space, pIoctl=0x%p sizeIn=%u\n", pIoctlStruct, sizeIn));
    dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    return -EFAULT;
  }

  spin_lock_irqsave(&pClObj->system.lock, f);
  if (NonBlockStateIdle != pClObj->system.nonblock.state) {
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    return -EBUSY;
  } else {
    pClObj->system.nonblock.state = NonBlockStateIoctl;
    pClObj->system.nonblock.pHead = pReq;
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    status = pHandler(pDevObj, pClObj, &pReq->clientRequest, reqCode, &pReq->ioctlStruct, sizeIn, sizeOut);
    spin_lock_irqsave(&pClObj->system.lock, f);
    if (DfIoStatusIsError(status)) {
      pClObj->system.nonblock.state = NonBlockStateIdle;
      pClObj->system.nonblock.pHead = NULL;
    } else {
      pClObj->system.nonblock.state = NonBlockStateDeferred;
    }
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    if (DfIoStatusIsError(status)) {
      dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    }
  }
  return status;
}

boolean_t
dfRequestClearCallback(
  DfClientRequest* pClientRequest)
{
  DfClientObject* pClObj;
  DfSpinLockFlags f;

  dfAssert(!pClientRequest->system.bCompleted);

  pClObj = pClientRequest->system.pClObj;

  spin_lock_irqsave(&pClObj->system.lock, f);
  if (pClientRequest->system.bCancel) {
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    return FALSE;
  } else {
    dfListRemove(&pClientRequest->system.node);
    pClientRequest->system.pCancelCallback = NULL;
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    return TRUE;
  }
}

void
dfRequestComplete(
  DfClientRequest* pClientRequest,
  DfIoStatus status)
{
  DfClientObject* pClObj;
  DfSpinLockFlags f;

  dfAssert(!pClientRequest->system.bCompleted);
  dfAssert(pClientRequest->system.pCancelCallback == NULL);

  pClObj = pClientRequest->system.pClObj;

  spin_lock_irqsave(&pClObj->system.lock, f);
  pClientRequest->system.status = status;
  pClientRequest->system.bCompleted = TRUE;
  if (NULL != pClientRequest->system.pWaitQueue) {
    wake_up(pClientRequest->system.pWaitQueue);
  }
  spin_unlock_irqrestore(&pClObj->system.lock, f);
}

boolean_t
dfRequestSetCallback(
  DfClientRequest* pClientRequest,
  DfCancelCallback* pCallback)
{
  DfClientObject* pClObj;
  DfSpinLockFlags f;

  dfAssert(!pClientRequest->system.bCompleted);
  dfAssert(pClientRequest->system.pCancelCallback == NULL);

  pClObj = pClientRequest->system.pClObj;

  spin_lock_irqsave(&pClObj->system.lock, f);
  if (pClientRequest->system.bCancel) {
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    return FALSE;
  } else {
    pClientRequest->system.pCancelCallback = pCallback;
    dfListAddToTail(&pClObj->system.requests.list, &pClientRequest->system.node);
    spin_unlock_irqrestore(&pClObj->system.lock, f);
    return TRUE;
  }
}

void
dfLinuxUserIfInit(
  void)
{
  sema_init(&g_vmaSemaphore, 1);
}

/*
** ---------------------------------------------------------------------------
** Linux driver framework special IOCTL handlers
** ---------------------------------------------------------------------------
*/

static int
ioctlFindMmapping(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  void* pIoctl)
{
  DfLinuxIoctlFindMmapping ioctl;
  DfMmappedRegion* pMapped;
  uint64_t virtualBase, virtualPfn;
  uint32_t index;
  boolean_t bByIndex;

  if (copy_from_user(&ioctl.in, pIoctl, sizeof(ioctl.in))) {
    dfDebugPrint(5, ("ioctlFindMmapping: fault copying ioctl struct from user space, pIoctl=0x%p size=%u\n",
      pIoctl, (unsigned int)sizeof(ioctl.in)));
    return -EFAULT;
  }

  bByIndex = (ioctl.in.bByIndex != 0);

  down(&g_vmaSemaphore);
  pMapped = pClObj->system.pMmappedRegions;
  if (bByIndex) {
    index = ioctl.in.variant.index;
    dfDebugPrint(5, ("ioctlFindMmapping: index=0x%lx\n", (unsigned long)index));
    while (0 != index && NULL != pMapped) {
      index--;
      pMapped = pMapped->system.pNext;
    }
  } else {
    virtualBase = ioctl.in.variant.virtualBase;
    virtualPfn = virtualBase >> PAGE_SHIFT;
    dfDebugPrint(5, ("ioctlFindMmapping: virtualBase=0x%08lx_%08lx\n", dfSplitUint64(virtualBase)));
    while (NULL != pMapped) {
      if ((pMapped->system.virtualBase >> PAGE_SHIFT) == virtualPfn) {
        break;
      }
      pMapped = pMapped->system.pNext;
    }
  }
  up(&g_vmaSemaphore);

  if (NULL == pMapped) {
    return DfLinuxMmapInvalidRegion;
  }

  ioctl.out.virtualBase = pMapped->system.virtualBase;
  ioctl.out.virtualLength = pMapped->system.virtualLength;
  dfDebugPrint(5, ("ioctlFindMmapping: found mapping, virtualBase=0x%llx virtualLength=0x%llx\n",
    (unsigned long long)ioctl.out.virtualBase, (unsigned long long)ioctl.out.virtualLength));

  if (copy_to_user(pIoctl, &ioctl.out, sizeof(ioctl.out))) {
    dfDebugPrint(5, ("ioctlFindMmapping: fault copying ioctl struct to user space, pIoctl=0x%p size=%u\n",
      pIoctl, (unsigned int)sizeof(ioctl.out)));
    return -EFAULT;
  }

  return DfLinuxMmapSuccess;
}

static int
ioctlGetMmapAddress(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  void* pIoctl)
{
  DfLinuxIoctlGetMmapAddress ioctl;
  DfMmappableRegion* pRegion;
  unsigned int regionTag;

  if (copy_from_user(&ioctl.in, pIoctl, sizeof(ioctl.in))) {
    dfDebugPrint(5, ("ioctlGetMmapAddress: fault copying ioctl struct from user space, pIoctl=0x%p size=%u\n",
      pIoctl, (unsigned int)sizeof(ioctl.in)));
    return -EFAULT;
  }

  regionTag = ioctl.in.regionTag;

  dfDebugPrint(5, ("ioctlGetMmapAddress: regionTag=%u(0x%x)\n", regionTag, regionTag));

  /* No existing mapping found, return info needed for mmap() from user-mode */
  pRegion = pDevObj->system.pMmappableRegions;
  while (NULL != pRegion) {
    if (pRegion->system.regionTag == ioctl.in.regionTag) {
      break;
    }
    pRegion = pRegion->system.pNext;
  }
  if (NULL == pRegion) {
    return DfLinuxMmapInvalidTag;
  }

  ioctl.out.address = (uint64_t)pRegion->system.pfnStart * PAGE_SIZE + pRegion->system.pageOffset;
  ioctl.out.length = pRegion->system.kernelSize;
  dfDebugPrint(5, ("ioctlGetMmapAddress: returning mmap info, address=0x%llx size=0x%llx\n",
    (unsigned long long)ioctl.out.address, (unsigned long long)ioctl.out.length));

  if (copy_to_user(pIoctl, &ioctl.out, sizeof(ioctl.out))) {
    dfDebugPrint(5, ("ioctlGetMmapAddress: fault copying ioctl struct to user space, pIoctl=0x%p size=%u\n",
      pIoctl, (unsigned int)sizeof(ioctl.out)));
    return -EFAULT;
  }

  return DfLinuxMmapSuccess;
}

static int
ioctlCancelRequests(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj)
{
  DfList complete;
  DfListNode* pHead;
  DfClientRequest* pCurrent;
  DfCancelCallback* pCancelCallback;
  DfSpinLockFlags f;

  dfDebugPrint(5, ("ioctlCancelRequests: entered\n"));

  /*
  ** Confiscate all current client requests but keep the list intact. Clears the
  ** list head pointer in the client object.
  */
  f = dfSpinLockGet(&pClObj->system.lock);
  dfListConfiscate(&complete, &pClObj->system.requests.list);
  dfSpinLockPut(&pClObj->system.lock, f);

  /* Now iterate through the confiscated list and cancel the client requests */
  while (1) {
    f = dfSpinLockGet(&pClObj->system.lock);
    pHead = dfListGetHead(&complete);
    if (pHead == NULL) {
      dfSpinLockPut(&pClObj->system.lock, f);
      break;
    }
    dfListRemove(pHead);
    pCurrent = DF_CONTAINER_OF(pHead, DfClientRequest, system.node);
    pCurrent->system.bCancel = TRUE;
    pCancelCallback = pCurrent->system.pCancelCallback;
    pCurrent->system.pCancelCallback = NULL;
    dfSpinLockPut(&pClObj->system.lock, f);      

    dfDebugPrint(2, ("ioctlCancelRequests: cancelling request 0x%p\n", pCurrent));
    dfAssert(NULL != pCancelCallback);
    pCancelCallback(pCurrent, DfIoStatusCancelled);
  }

  return DfLinuxCancelSuccess;
}

static int
ioctlFinishNonBlock(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  void* pIoctlOut,
  unsigned int sizeOut)
{
  DfIoStatus status;
  DfLinuxClientRequestNonBlock* pReq;
  DfSpinLockFlags f;

  f = dfSpinLockGet(&pClObj->system.lock);
  switch (pClObj->system.nonblock.state) {
  case NonBlockStateDeferred:
    pReq = pClObj->system.nonblock.pHead;
    dfAssert(NULL != pReq);
    if (pReq->clientRequest.system.bCompleted) {
      pClObj->system.nonblock.state = NonBlockStateIdle;
      pClObj->system.nonblock.pHead = NULL;
      dfSpinLockPut(&pClObj->system.lock, f);
      status = pReq->clientRequest.system.status;
      if (status == DfIoStatusSuccess && 0 != sizeOut) {
        if (copy_to_user(pIoctlOut, &pReq->ioctlStruct, sizeOut)) {
          dfDebugPrint(5, ("ioctlFinishNonBlocking: fault copying ioctl struct to user space, pIoctlOut=0x%p sizeOut=%u\n",
                           pIoctlOut, sizeOut));
          status = -EFAULT;
        }
      }
      dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    } else {
      dfSpinLockPut(&pClObj->system.lock, f);
      status = -EINPROGRESS;
    }
    break;

  case NonBlockStateIdle:
  case NonBlockStateIoctl:
  default:
    status = -EPIPE;
    dfSpinLockPut(&pClObj->system.lock, f);
    break;

  }

  return status;
}

static int
specialIoctl(
  DfDeviceObject* pDevObj,
  DfClientObject* pClObj,
  unsigned int code,
  void* pIoctl,
  unsigned int sizeOut)
{
  dfDebugPrint(5, ("specialIoctl: entered, code=%u(0x%x) sizeOut=%u pIoctl=%p\n", code, code, sizeOut, pIoctl));

  switch (code) {
  case DFLINUX_IOCTLCODE_FINDMMAPPING:
    return ioctlFindMmapping(pDevObj, pClObj, pIoctl);
    break;

  case DFLINUX_IOCTLCODE_GETMMAPADDRESS:
    return ioctlGetMmapAddress(pDevObj, pClObj, pIoctl);
    break;

  case DFLINUX_IOCTLCODE_CANCEL:
    return ioctlCancelRequests(pDevObj, pClObj);
    break;

  case DFLINUX_IOCTLCODE_FINISH:
    return ioctlFinishNonBlock(pDevObj, pClObj, pIoctl, sizeOut);
    break;

  default:
    return -EINVAL;
  }
}

/*
** ---------------------------------------------------------------------------
** file_operations methods
**
** These functions are called for every DfInterface object, regardless of
** class. In general, after identifying the DfInterface object associated with
** the 'struct file' object, they call the methods associated with the
** DfInterface object in order to achieve the desired class-specific behavior.
** ---------------------------------------------------------------------------
*/

static int
fopsOpen(
  struct inode* pInode,
  struct file* pFile)
{
  DfOpenStatus err = 0;
  struct cdev* pCdev;
  DfInterface* pInterface;
  DfDeviceObject* pDevObj;
  DfClientObject* pClObj;
  DfOnOpenMethod* pOnOpen;
  boolean_t bPassive = TRUE;

  dfDebugPrint(3, ("fopsOpen: entered, pInode=0x%p pFile=0x%p i_mode=0x%lx f_mode=0x%lx\n",
    pInode, pFile, (unsigned long)pInode->i_mode, (unsigned long)pFile->f_mode));

  pCdev = pInode->i_cdev;
  pInterface = DF_CONTAINER_OF(pCdev, DfInterface, system.cdev);
  pDevObj = pInterface->system.pDevObj;

  if ((pFile->f_mode & FMODE_WRITE) == FMODE_WRITE) {
    if (capable(CAP_SYS_ADMIN)) {
      /* User is adminstrator - access all areas */
      bPassive = FALSE;
    } else {
      /* Reject non-administrator */
      return -EACCES;
    }
  } else {
    /* "Read-only" mode - IOCTLs that attempt to alter state of device will fail */
    bPassive = TRUE;
  }

  pClObj = (DfClientObject*)dfKMalloc(sizeof(*pClObj), GFP_KERNEL);
  if (pClObj == NULL) {
    return -ENOMEM;
  }
  dfSpinLockInit(&pClObj->system.lock);
  pClObj->system.pFile = pFile;
  pClObj->system.pDevObj = pDevObj;
  pClObj->system.pInterface = pInterface;
  pClObj->system.bPassive = bPassive;
  pClObj->system.pMmappedRegions = NULL;
  pClObj->system.cleanup.state = ClientStateOpen;
  pClObj->system.requests.count = 0;
  dfListInit(&pClObj->system.requests.list);
  pClObj->system.nonblock.state = NonBlockStateIdle;
  pClObj->system.nonblock.pHead = NULL;
  init_waitqueue_head(&pClObj->system.nonblock.pollWaitQueue);
  pClObj->system.pContext = NULL;
  pFile->private_data = (void*)pClObj;

  pOnOpen = pInterface->system.methods.pOnOpen;
  if (NULL != pOnOpen) {
    err = pOnOpen(pDevObj, pClObj, pInterface, bPassive ? FALSE : TRUE);
    if (err) {
      dfFree(pClObj);
      return err;
    }
  }

  nonseekable_open(pInode, pFile);

  return 0; /* Success */
}

static int
fopsRelease(
  struct inode* pInode,
  struct file* pFile)
{
  DfDeviceObject* pDevObj;
  DfClientObject* pClObj;
  DfInterface* pInterface;
  DfOnCleanupMethod* pOnCleanup;
  DfOnCloseMethod* pOnClose;
  DfSpinLockFlags f;
  DfLinuxClientRequestNonBlock* pReq = NULL; /* init to prevent compiler whine */
  DfCancelCallback* pCancelCallback = NULL;
  boolean_t bMustWait = FALSE;
  int err = 0;
  
  dfDebugPrint(3, ("fopsRelease: entered, pInode=0x%p pFile=0x%p\n", pInode, pFile));

  pClObj = (DfClientObject*)pFile->private_data;
  dfAssert(NULL != pClObj);

  pInterface = pClObj->system.pInterface;
  pDevObj = pClObj->system.pDevObj;

  f = dfSpinLockGet(&pClObj->system.lock);
  if (pClObj->system.nonblock.state != NonBlockStateIdle) {
    dfAssert(pClObj->system.nonblock.state >= NonBlockStateDeferred);
    pReq = pClObj->system.nonblock.pHead;
    dfAssert(NULL != pReq);
    if (!pReq->clientRequest.system.bCompleted) {
      dfAssert(!pReq->clientRequest.system.bCancel);
      pReq->clientRequest.system.bCancel = TRUE;
      pCancelCallback = pReq->clientRequest.system.pCancelCallback;
      if (NULL != pCancelCallback) {
        pReq->clientRequest.system.pCancelCallback = NULL;
        dfListRemove(&pReq->clientRequest.system.node);
      }
      bMustWait = TRUE;
    }
  }
  dfSpinLockPut(&pClObj->system.lock, f);
  if (NULL != pCancelCallback) {
    pCancelCallback(&pReq->clientRequest, DfIoStatusCancelled);
  }
  
  pOnCleanup = pInterface->system.methods.pOnCleanup;
  if (NULL != pOnCleanup) {
    pOnCleanup(pDevObj, pClObj);
  }

  if (bMustWait) {
    requestWait(&pReq->clientRequest);
    dfPoolFree(&pDevObj->system.request.poolNonBlock, pReq);
    pClObj->system.nonblock.pHead = NULL;
    pClObj->system.nonblock.state = NonBlockStateIdle;
  }

  pOnClose = pInterface->system.methods.pOnClose;
  if (NULL != pOnClose) {
    pOnClose(pDevObj, pClObj);
  }

  /* There should be no regions still mapped */
  dfAssert(pClObj->system.pMmappedRegions == NULL);

  dfFree(pClObj);

  return err;
}

#if _IOC_SIZEBITS < 12
# error _IOC_SIZEBITS is less than 12
#endif

static int
fopsIoctl(
  struct inode* pInode,
  struct file* pFile,
  unsigned int code,
  unsigned long arg)
{
  DfClientObject* pClObj;
  DfDeviceObject* pDevObj;
  DfInterface* pInterface;
  DfOnIoctlMethod* pHandler;
  DfIoStatus status = DfIoStatusInvalidCode;
  unsigned int iocSize, sizeIn, sizeOut, reqCode;
  boolean_t bNonBlock;

  dfDebugPrint(5, ("fopsIoctl: entered, pInode=0x%p pFile=0x%p code=%u(%x) arg=%lu(0x%lx)\n",
		   pInode, pFile, code, code, arg, arg));

  pClObj = (DfClientObject*)pFile->private_data;
  dfAssert(NULL != pClObj);
  pDevObj = pClObj->system.pDevObj;
  pInterface = pClObj->system.pInterface;

  /* The IOCTL structure in and out sizes are encoded in the IOC size field of 'code' */
  iocSize = _IOC_SIZE(code);
  sizeIn = iocSize & 0x3fU;
  sizeOut = (iocSize >> 6) & 0x3fU;
  reqCode = _IOC_NR(code);
  if (reqCode & DFLINUX_IOCTL_BASE) {
    return specialIoctl(pDevObj, pClObj, reqCode, (void*)(size_t)arg, sizeOut);
  }
  bNonBlock = (reqCode & 0x40U) ? TRUE : FALSE;
  reqCode &= 0x3fU;
  dfAssert(!bNonBlock || 0 == sizeOut);

  dfDebugPrint(5, ("fopsIoctl: sizeIn=%u sizeOut=%u reqCode=%u bNonBlock=%s\n",
    sizeIn, sizeOut, reqCode, bNonBlock ? "TRUE" : "FALSE"));

  pHandler = pInterface->system.methods.pOnIoctl;
  if (NULL != pHandler) {
    if (bNonBlock) {
      status = deliverRequestNonBlocking(pDevObj, pClObj, pHandler, reqCode, (void*)(uintptr_t)arg, sizeIn, sizeOut);
    } else {
      status = deliverRequestBlocking(pDevObj, pClObj, pHandler, reqCode, (void*)(uintptr_t)arg, sizeIn, sizeOut);
    }
  }

  return status;
}

static long
fopsUnlockedIoctl(
  struct file* pFile,
  unsigned int code,
  unsigned long arg)
{
  dfDebugPrint(6, ("fopsUnlockedIoctl: entered, pFile=0x%p code=%u(%x) arg=%lu(0x%lx)\n", pFile, code, code, arg, arg));

  return fopsIoctl(pFile->f_dentry->d_inode, pFile, code, arg);
}

#if DF_NEED_THUNK
static long
fopsCompatIoctl(
  struct file* pFile,
  unsigned int code,
  unsigned long arg)
{
  DfClientObject* pClObj;
  DfDeviceObject* pDevObj;
  DfInterface* pInterface;
  void* pArg;
  DfOnIoctlMethod* pHandler;
  DfIoStatus status = DfIoStatusInvalidCode;
  unsigned int iocSize, sizeIn, sizeOut, reqCode;
  boolean_t bNonBlock;

  dfDebugPrint(5, ("fopsCompatIoctl: entered, pFile=0x%p code=%u(%x) arg=%lu(0x%lx)\n",
		   pFile, code, code, arg, arg));

  pClObj = (DfClientObject*)pFile->private_data;
  dfAssert(NULL != pClObj);
  pDevObj = pClObj->system.pDevObj;
  pInterface = pClObj->system.pInterface;
  dfAssert(NULL != pClObj);

  pArg = compat_ptr(arg);

  /* The IOCTL structure in and out sizes are encoded in the IOC size field of 'code' */
  iocSize = _IOC_SIZE(code);
  sizeIn = iocSize & 0x3fU;
  sizeOut = (iocSize >> 6) & 0x3fU;
  reqCode = _IOC_NR(code);
  if (reqCode & DFLINUX_IOCTL_BASE) {
    return specialIoctl(pDevObj, pClObj, reqCode, pArg, sizeOut);
  }
  bNonBlock = (reqCode & 0x40U) ? TRUE : FALSE;
  reqCode &= 0x3fU;
  dfAssert(!bNonBlock || 0 == sizeOut);

  dfDebugPrint(5, ("fopsCompatIoctl: sizeIn=%u sizeOut=%u reqCode=%u bNonBlock=%s\n",
    sizeIn, sizeOut, reqCode, bNonBlock ? "TRUE" : "FALSE"));

  pHandler = pInterface->system.methods.pOnIoctlThunk;
  if (NULL != pHandler) {
    if (bNonBlock) {
      status = deliverRequestNonBlocking(pDevObj, pClObj, pHandler, reqCode, pArg, sizeIn, sizeOut);
    } else {
      status = deliverRequestBlocking(pDevObj, pClObj, pHandler, reqCode, pArg, sizeIn, sizeOut);
    }
  }

  return status;
}
#endif

static void
vmaClose(
  struct vm_area_struct* pVma)
{
  DfClientObject* pClObj;
  DfDeviceObject* pDevObj;
  DfMmappedRegion* pMapped;
  DfMmappedRegion** ppMapped;

  dfDebugPrint(3, ("vmaClose: entered, pVma=0x%p pVma->vm_private_data=0x%p\n", pVma, pVma->vm_private_data));

  dfAssert(pVma->vm_file != NULL);
  dfAssert(pVma->vm_private_data != NULL);

  /* Remove corresponding DfMmappedRegion from client object, and then free it */
  down(&g_vmaSemaphore);
  pMapped = (DfMmappedRegion*)pVma->vm_private_data;
  pClObj = (DfClientObject*)pMapped->system.pClObj;
  pDevObj = (DfDeviceObject*)pMapped->system.pDevObj;
  dfAssert(NULL != pDevObj);
  if (NULL != pClObj) {
    ppMapped = &pClObj->system.pMmappedRegions;
    pMapped = *ppMapped;
    while (NULL != pMapped) {
      if (pMapped->system.pVma == pVma) {
        break;
      }
      ppMapped = &pMapped->system.pNext;
      pMapped = *ppMapped;
    }
    dfAssert(NULL != pMapped);
    if (NULL != pMapped) {
      *ppMapped = pMapped->system.pNext;
    }
  }
  pMapped->system.count--;
  if (pMapped->system.count == 0) {
    /* Decrement reference count of device object */
    kobject_put(&pDevObj->system.device.pPci->dev.kobj);
    kfree(pMapped);
  }
  up(&g_vmaSemaphore);
}

/* We define this function to prevent mremap from being able to extend an existing mapping */
#if NO_PAGE_STRUCT_VM_FAULT
static int
vmaFault(
  struct vm_area_struct* pVma,
  struct vm_fault* pVmf)
{
  dfDebugPrint(6, ("vmaFault: entered, pVma=0x%p pVma->vm_private_data=0x%p\n", pVma, pVma->vm_private_data));

  dfAssert(pVma->vm_file != NULL);
  dfAssert(pVma->vm_private_data != NULL);

  return VM_FAULT_SIGBUS;
}
#else
static struct page*
vmaNoPage(
  struct vm_area_struct* pVma,
  unsigned long address,
  int* pType)
{
  dfDebugPrint(6, ("vmaNoPage: entered, pVma=0x%p address=0x%lx pVma->vm_private_data=0x%p\n",
		   pVma, address, pVma->vm_private_data));

  dfAssert(pVma->vm_file != NULL);
  dfAssert(pVma->vm_private_data != NULL);

  return NOPAGE_SIGBUS;
}
#endif

static void
vmaOpen(
  struct vm_area_struct* pVma)
{
  DfMmappedRegion* pMapped;

  dfDebugPrint(3, ("vmaOpen: entered, pVma=0x%p pVma->vm_private_data=0x%p\n", pVma, pVma->vm_private_data));

  dfAssert(pVma->vm_file != NULL);
  dfAssert(pVma->vm_private_data != NULL);

  pMapped = (DfMmappedRegion*)pVma->vm_private_data;

  down(&g_vmaSemaphore);
  pMapped->system.count++;
  up(&g_vmaSemaphore);
}

static struct vm_operations_struct vmOps = {
  .open = vmaOpen,
  .close = vmaClose,
#if NO_PAGE_STRUCT_VM_FAULT
  .fault = vmaFault
#else
  .nopage = vmaNoPage
#endif
};

static int
fopsMmap(
  struct file* pFile,
  struct vm_area_struct* pVma)
{
  DfDeviceObject* pDevObj;
  DfClientObject* pClObj;
  DfMmappableRegion* pRegion;
  DfMmappedRegion* pMapped;
  unsigned long size, numPage;

  dfDebugPrint(3, ("fopsMmap: entered, pFile=0x%p pVma=0x%p\n", pFile, pVma));

  pClObj = (DfClientObject*)pFile->private_data;
  dfAssert(NULL != pClObj);
  pDevObj = pClObj->system.pDevObj;

  if (pClObj->system.bPassive) {
    return -EPERM; /* Don't let passively opened handles do mmap() */
  }

  size = pVma->vm_end - pVma->vm_start;
  numPage = (size + PAGE_SIZE - 1) >> PAGE_SHIFT;
  if (0 == numPage) {
    return -EINVAL; /* sanity check */
  }

  pRegion = pDevObj->system.pMmappableRegions;
  while (NULL != pRegion) {
    if (pVma->vm_pgoff >= pRegion->system.pfnStart && pVma->vm_pgoff <= pRegion->system.pfnEnd) {
      break;
    }
    pRegion = pRegion->system.pNext;
  }
  if (NULL == pRegion) {
    return -EINVAL;
  }
  if (pVma->vm_pgoff + numPage - 1 > pRegion->system.pfnEnd) {
    return -EINVAL; /* attempting to map past end of region */
  }

  pMapped = dfKMalloc(sizeof(*pMapped), GFP_KERNEL);
  if (NULL == pMapped) {
    return -ENOMEM;
  }
  pMapped->system.pDevObj = pDevObj;
  pMapped->system.pClObj = pClObj;
  pMapped->system.pNext = NULL;
  pMapped->system.pRegion = pRegion;
  pMapped->system.virtualBase = pVma->vm_start;
  pMapped->system.virtualLength = size;
  pMapped->system.pVma = pVma;
  pMapped->system.count = 1;

  down(&g_vmaSemaphore);
  pVma->vm_page_prot = pgprot_noncached(pVma->vm_page_prot);
  pVma->vm_flags |= VM_RESERVED;  /* Don't swap physical pages */
  pVma->vm_flags |= VM_IO;	  /* Don't core dump IO pages */
  if (io_remap_pfn_range(pVma, pVma->vm_start, pVma->vm_pgoff, size, pVma->vm_page_prot)) {
    dfDebugPrint(0, ("*** fopsMmap: io_remap_pfn_range failed, pVma=0x%p pVma->vm_start=0x%p pfn=%lu(0x%lx), size=%lu(0x%lx)\n",
      pVma, (void*)(size_t)pVma->vm_start, (unsigned long)pVma->vm_pgoff, (unsigned long)pVma->vm_pgoff,
      (unsigned long)size, (unsigned long)size));
    up(&g_vmaSemaphore);
    dfFree(pMapped);
    return -EAGAIN;
  }
  pMapped->system.pNext = pClObj->system.pMmappedRegions;
  pClObj->system.pMmappedRegions = pMapped;
  pVma->vm_private_data = (void*)pMapped;
  pVma->vm_ops = &vmOps;
  /* Increment reference count of device object */
  kobject_get(&pDevObj->system.device.pPci->dev.kobj);
  up(&g_vmaSemaphore);

  return 0;
}

static unsigned int
fopsPoll(
  struct file* pFile,
  struct poll_table_struct* pTable)
{
  DfClientObject* pClObj;
  DfLinuxClientRequestNonBlock* pReq;
  DfSpinLockFlags f;
  unsigned int mask = 0;

  dfDebugPrint(5, ("fopsPoll: entered, pFile=0x%p pTable=0x%p\n", pFile, pTable));

  pClObj = (DfClientObject*)pFile->private_data;
  dfAssert(NULL != pClObj);

  f = dfSpinLockGet(&pClObj->system.lock);
  switch (pClObj->system.nonblock.state) {
  case NonBlockStateIoctl:
  case NonBlockStateDeferred:
    pReq = pClObj->system.nonblock.pHead;
    dfAssert(NULL != pReq);
    if (pReq->clientRequest.system.bCompleted) {
      dfSpinLockPut(&pClObj->system.lock, f);
      mask |= POLLPRI;
    } else { 
      pReq->clientRequest.system.pWaitQueue = &pClObj->system.nonblock.pollWaitQueue;
      poll_wait(pFile, &pClObj->system.nonblock.pollWaitQueue, pTable);
      dfSpinLockPut(&pClObj->system.lock, f);
    }
    break;

  case NonBlockStateIdle:
    dfSpinLockPut(&pClObj->system.lock, f);
    break;

  default:
    dfAssert(FALSE);
    dfSpinLockPut(&pClObj->system.lock, f);
    break;
  }

  return mask;
}

static struct file_operations g_fops = {
  .owner = THIS_MODULE,
#if DF_NEED_THUNK
  .compat_ioctl = fopsCompatIoctl,
#endif
  .ioctl = fopsIoctl,
  .llseek = no_llseek,
  .mmap = fopsMmap,
  .open = fopsOpen,
  .poll = fopsPoll,
  .release = fopsRelease,
  .unlocked_ioctl = fopsUnlockedIoctl
};

/*
** -----------------------------------------------------------------
** Functions exported by Linux driver framework
** -----------------------------------------------------------------
*/

boolean_t
dfInterfaceClassRegister(
  DfDriverObject* pDrvObj,
  const char* pClassName,
  DfInterfaceClass** ppClass)
{
  DfInterfaceClass* pClass;
  dev_t devNum;
  int result;

  pClass = (DfInterfaceClass*)dfKMalloc(sizeof(*pClass), GFP_KERNEL);
  if (NULL == pClass) {
    dfDebugPrint(0, ("*** dfInterfaceClassRegister: failed to allocate struct, pDrvObj=0x%p\n", pDrvObj));
    return FALSE;
  }
  dfStrCpy(pClass->system.name, DF_ARRAY_LENGTH(pClass->system.name), pClassName);
#if USE_CLASS_SIMPLE
  pClass->system.pClass = class_simple_create(THIS_MODULE, (char*)pClassName);
#else
  pClass->system.pClass = class_create(THIS_MODULE, pClassName);
#endif
  if (IS_ERR(pClass->system.pClass)) {
    kfree(pClass);
    return FALSE;
  }
  result = alloc_chrdev_region(&devNum, 0, MaxNumMinor, pClassName); /* This should be in interface class */
  if (result < 0) {
#if USE_CLASS_SIMPLE
    class_simple_destroy(pClass->system.pClass);
#else
    class_destroy(pClass->system.pClass);
#endif
    kfree(pClass);
    return FALSE;
  }
  pClass->system.devNum = devNum;
  pClass->system.numInterface = 0;

  *ppClass = pClass;
  return TRUE;
}

void
dfInterfaceClassUnregister(
  DfInterfaceClass* pClass)
{
  dfAssert(NULL != pClass);
  dfAssert(NULL != pClass && 0 == pClass->system.numInterface);
  unregister_chrdev_region(pClass->system.devNum, MaxNumMinor);
#if USE_CLASS_SIMPLE
  class_simple_destroy(pClass->system.pClass);
#else
  class_destroy(pClass->system.pClass);
#endif
  kfree(pClass);
}

DF_ATTR_THREAD_CONTEXT
boolean_t
dfInterfaceRegister(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface,
  DfInterfaceClass* pClass,
  const DfInterfaceMethods* pMethods)
{
  int index;

  pInterface->system.methods.pOnCleanup = pMethods->pOnCleanup;
  pInterface->system.methods.pOnClose = pMethods->pOnClose;
  pInterface->system.methods.pOnIoctl = pMethods->pOnIoctl;
#if DF_NEED_THUNK
  pInterface->system.methods.pOnIoctlThunk = pMethods->pOnIoctlThunk;
#endif
  pInterface->system.methods.pOnOpen = pMethods->pOnOpen;
  pInterface->system.pDevObj = pDevObj;
  pInterface->system.pNext = NULL;
  pInterface->system.pContext = NULL;
  pInterface->system.bEnabled = FALSE;
  pDevObj->system.interfaceList.pHead = pInterface;
  pInterface->system.pClass = pClass;
  cdev_init(&pInterface->system.cdev, &g_fops);
  pInterface->system.cdev.owner = THIS_MODULE;
  index = pClass->system.numInterface;
  pInterface->system.index = index;
  pClass->system.numInterface = index + 1;
  return TRUE;
}

DF_ATTR_THREAD_CONTEXT
boolean_t
dfInterfaceUnregister(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface)
{
  DfInterfaceClass* pClass;
  DfInterface** pp;
  DfInterface* p;

  dfAssert(!pInterface->system.bEnabled);
  dfAssert(NULL != pInterface->system.pClass);

  pClass = pInterface->system.pClass;
  pClass->system.numInterface--;

  pp = &pDevObj->system.interfaceList.pHead;
  while (1) {
    p = *pp;
    if (NULL == p) {
      break;
    }
    if (p == pInterface) {
      *pp = p->system.pNext;
      break;
    }
  }

  return (NULL == p) ? FALSE : TRUE;
}

DF_ATTR_THREAD_CONTEXT
boolean_t
dfInterfaceEnable(
  DfDeviceObject* pDevObj,
  DfInterface* pInterface,
  boolean_t bEnable)
{
  DfInterfaceClass* pClass;
  dev_t devNum;
  int result;

  pClass = pInterface->system.pClass;
  devNum = MKDEV(MAJOR(pClass->system.devNum), MINOR(pClass->system.devNum) + pInterface->system.index);
  if (bEnable) {
    dfAssert(!pInterface->system.bEnabled);
    if (pInterface->system.index >= MaxNumMinor) {
      dfDebugPrint(0, ("*** dfInterfaceEnable: can't enable; minor numbers exhausted, index=%u\n", pInterface->system.index));
      return FALSE;
    }
    result = cdev_add(&pInterface->system.cdev, devNum, 1);
    if (result < 0) {
      return FALSE;
    }
#if USE_CLASS_SIMPLE
    pInterface->system.pDevice = class_simple_device_add(pClass->system.pClass, devNum, &pDevObj->system.device.pPci->dev, "%s%d", pClass->system.name, pInterface->system.index);
#else
    pInterface->system.pDevice = device_create(pClass->system.pClass, &pDevObj->system.device.pPci->dev, devNum, pDevObj, "%s%d", pClass->system.name, pInterface->system.index);
#endif
    if (NULL == pInterface->system.pDevice) {
      cdev_del(&pInterface->system.cdev);
      return FALSE;
    }
    pInterface->system.bEnabled = TRUE;
    return TRUE;
  } else {
    dfAssert(pInterface->system.bEnabled);
    pInterface->system.bEnabled = FALSE;
#if USE_CLASS_SIMPLE
    class_simple_device_remove(devNum);
#else
    device_destroy(pClass->system.pClass, devNum);
#endif
    cdev_del(&pInterface->system.cdev);
    return TRUE;
  }
}

