/*
** File: flash.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for Flash programming.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include "device.h"
#include "flash.h"
#include <admxrc3/types.h>

#define READFLASH_BUFFER_SIZE  (0x100)
#define WRITEFLASH_BUFFER_SIZE (0x100)

typedef struct _Context {
  boolean_t bCancel;
  DfIoStatus reason;
  DfEvent event;
} Context;

static DfIoStatus
mapAcquireStatus(
  CoreAcquireStatus acqStatus)
{
  switch (acqStatus) {
  case CoreAcquireSuccess:
    return DfIoStatusSuccess;

  case CoreAcquireMustWait:
    return DfIoStatusError(ADMXRC3_MUST_WAIT);

  case CoreAcquireCancelled:
    return DfIoStatusError(ADMXRC3_CANCELLED);

  case CoreAcquireRemoved:
  case CoreAcquireFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);

  case CoreAcquireInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);

  case CoreAcquireInvalid:
  case CoreAcquireUnexpectedError:
  case CoreAcquireWaiting:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static DfIoStatus
mapFlashStatus(
  CoreFlashStatus flashStatus)
{
  switch (flashStatus) {
  case CoreFlashSuccess:
    return DfIoStatusSuccess;
  case CoreFlashInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  case CoreFlashNotPresent:
  case CoreFlashTimeout:
  case CoreFlashVerifyFailed:
    return DfIoStatusError(ADMXRC3_HARDWARE_ERROR);
  case CoreFlashInvalidRegion:
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  case CoreFlashInterrupted:
    return DfIoStatusError(ADMXRC3_CANCELLED);
  case CoreFlashInvalidParameter:
  case CoreFlashGeneralFailure:
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR);
  }
}

static void
cancelCallback(
  DfClientRequest* pReq,
  DfIoStatus reason)
{
  Context* pContext = (Context*)dfRequestGetContext(pReq);

  pContext->reason = reason;
  pContext->bCancel = TRUE;
  dfEventSignal(&pContext->event);
}

DfIoStatus
ioctlEraseFlash(
  Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = ADMXRC3_FLASH_SYNC;
#endif
  IOCTLS_ADMXRC3_ERASEFLASH* pIoctl = (IOCTLS_ADMXRC3_ERASEFLASH*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreAcquireStatus acqStatus;
  CoreFlashStatus flashStatus;
  CoreTicket ticket;
  unsigned int bankIndex;
  uint32_t flags;
  uint64_t offset, length;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }

  bankIndex = pIoctl->in.flashIndex;
  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  dfDebugPrint(3, ("ioctlEraseFlash: bankIndex=%lu flags=0x%lx offset=%llu(0x%08lx_%08lx) length=%llu(0x%08lx_%08lx)\n",
    (unsigned long)bankIndex, (unsigned long)flags, (unsigned long long)offset, dfSplitUint64(offset),
    (unsigned long long)length, dfSplitUint64(length)));

  if (bankIndex >= pDevCtx->info.numFlashBank) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  ticket.count = 1;
  ticket.resources[0].code = CORE_RESOURCE_FLASH;
  ticket.resources[0].index = (uint8_t)bankIndex;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &ticket);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &ticket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }

  flashStatus = pDevCtx->coreInterface.pFlashErase(pDevCtx->pCoreContext, &ticket, bankIndex, offset, length);
  if (CoreFlashSuccess != flashStatus) {
    status = mapFlashStatus(flashStatus);
  }

  pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &ticket);
  return status;
}

DfIoStatus
ioctlGetFlashInfo(
  Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETFLASHINFO* pIoctl = (IOCTLS_ADMXRC3_GETFLASHINFO*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreFlashStatus flashStatus;
  CoreFlashInfo flashInfo;
  unsigned int bankIndex;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize) {
    return DfIoStatusInvalid;
  }

  bankIndex = pIoctl->in.flashIndex;

  dfDebugPrint(3, ("ioctlGetFlashInfo: bankIndex=%lu\n", (unsigned long)bankIndex));

  flashStatus = pDevCtx->coreInterface.pGetFlashInfo(pDevCtx->pCoreContext, bankIndex, &flashInfo);
  if (CoreFlashSuccess == flashStatus) {
    pIoctl->out.size = flashInfo.size;
    pIoctl->out.useableStart = flashInfo.useableStart;
    pIoctl->out.useableLength = flashInfo.useableLength;
    pIoctl->out.vendorId = flashInfo.vendorId;
    pIoctl->out.deviceId = flashInfo.deviceId;
  } else {
    status = mapFlashStatus(flashStatus);
  }

  return status;
}

DfIoStatus
ioctlGetFlashBlockInfo(
  Admxrc3DeviceContext* pDevCtx,
  void* pBuffer,
  unsigned int inSize,
  unsigned int outSize)
{
  IOCTLS_ADMXRC3_GETFLASHBLOCKINFO* pIoctl = (IOCTLS_ADMXRC3_GETFLASHBLOCKINFO*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreFlashStatus flashStatus;
  CoreFlashBlockInfo blockInfo;
  unsigned int bankIndex;
  uint64_t address;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize || sizeof(pIoctl->out) != outSize) {
    return DfIoStatusInvalid;
  }

  bankIndex = pIoctl->in.flashIndex;
  address = pIoctl->in.address;

  dfDebugPrint(3, ("ioctlGetFlashBlockInfo: bankIndex=%lu address=%llu(0x%08lx_%08lx)\n",
    (unsigned long)bankIndex, (unsigned long long)address, dfSplitUint64(address)));

  flashStatus = pDevCtx->coreInterface.pGetFlashBlockInfo(pDevCtx->pCoreContext, bankIndex, address, &blockInfo);
  if (CoreFlashSuccess == flashStatus) {
    pIoctl->out.address = blockInfo.address;
    pIoctl->out.length = blockInfo.length;
    pIoctl->out.flags = blockInfo.flags;
  } else {
    status = mapFlashStatus(flashStatus);
  }

  return status;
}

DfIoStatus
ioctlReadFlash(
  Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = ADMXRC3_FLASH_SYNC;
#endif
  IOCTLS_ADMXRC3_READFLASH* pIoctl = (IOCTLS_ADMXRC3_READFLASH*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreAcquireStatus acqStatus;
  CoreFlashStatus flashStatus;
  CoreFlashInfo flashInfo;
  CoreTicket ticket;
  uint8_t buffer[READFLASH_BUFFER_SIZE];
  unsigned int bankIndex;
  uint32_t flags;
  uint64_t offset;
  size_t length, chunk;
  uint8_t* pUserBuf;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  bankIndex = pIoctl->in.flashIndex;
  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  pUserBuf = (uint8_t*)pIoctl->in.pBuffer;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  dfDebugPrint(3, ("ioctlReadFlash: bankIndex=%lu flags=0x%lx offset=%llu(0x%08lx_%08lx) length=%llu(0x%08lx_%08lx) pUserBuf=0x%p\n",
    (unsigned long)bankIndex, (unsigned long)flags, (unsigned long long)offset, dfSplitUint64(offset),
    (unsigned long long)length, dfSplitUint64((uint64_t)length), pUserBuf));

  if (bankIndex >= pDevCtx->info.numFlashBank) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  flashStatus = pDevCtx->coreInterface.pGetFlashInfo(pDevCtx->pCoreContext, bankIndex, &flashInfo);
  if (CoreFlashSuccess != flashStatus) {
    status = mapFlashStatus(flashStatus);
    goto done;
  }
  if (offset + length < offset) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }
  if (offset < flashInfo.useableStart) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }
  if (offset + length > flashInfo.useableStart + flashInfo.useableLength) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }

  ticket.count = 1;
  ticket.resources[0].code = CORE_RESOURCE_FLASH;
  ticket.resources[0].index = (uint8_t)bankIndex;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &ticket);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &ticket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuf, length, TRUE);

  while (length) {
    chunk = (length > sizeof(buffer)) ? sizeof(buffer) : length;
    flashStatus = pDevCtx->coreInterface.pFlashRead(pDevCtx->pCoreContext, &ticket, bankIndex, offset, chunk, buffer);
    if (CoreFlashSuccess != flashStatus) {
      status = mapFlashStatus(flashStatus);
      goto done;
    }
    dfCopyKU(pUserBuf, buffer, chunk);
    offset += chunk;
    pUserBuf += chunk;
    length -= chunk;
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

done:
  pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &ticket);
  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlReadFlashThunk(
  Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_READFLASH* pIoctl32 = (IOCTLS32_ADMXRC3_READFLASH*)pBuffer;
  IOCTLS_ADMXRC3_READFLASH ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.flashIndex = pIoctl32->in.flashIndex;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.offset = pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.pBuffer = dfThunkPtr(pIoctl32->in.pBuffer);
  return ioctlReadFlash(pDevCtx, pClCtx, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlWriteFlash(
  Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = ADMXRC3_FLASH_SYNC;
#endif
  IOCTLS_ADMXRC3_WRITEFLASH* pIoctl = (IOCTLS_ADMXRC3_WRITEFLASH*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreAcquireStatus acqStatus;
  CoreFlashStatus flashStatus;
  CoreFlashInfo flashInfo;
  CoreFlashBlockInfo blockInfo;
  CoreTicket ticket;
  Context context;
  unsigned int bankIndex;
  uint32_t flags;
  uint64_t offset, blockOffset, chunk;
  size_t length;
  const uint8_t* pUserBuf;
  void* pStagingBuffer;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  bankIndex = pIoctl->in.flashIndex;
  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  pUserBuf = (const uint8_t*)pIoctl->in.pBuffer;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  dfDebugPrint(3, ("ioctlWriteFlash: bankIndex=%lu flags=0x%lx offset=%llu(0x%08lx_%08lx) length=%llu(0x%08lx_%08lx), pUserBuf=0x%p\n",
    (unsigned long)bankIndex, (unsigned long)flags, (unsigned long long)offset, dfSplitUint64(offset),
    (unsigned long long)length, dfSplitUint64((uint64_t)length), pUserBuf));

  if (bankIndex >= pDevCtx->info.numFlashBank) {
    return DfIoStatusError(ADMXRC3_INVALID_INDEX);
  }

  flashStatus = pDevCtx->coreInterface.pGetFlashInfo(pDevCtx->pCoreContext, bankIndex, &flashInfo);
  if (CoreFlashSuccess != flashStatus) {
    status = mapFlashStatus(flashStatus);
    goto done;
  }
  if (offset + length < offset) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }
  if (offset < flashInfo.useableStart) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }
  if (offset + length > flashInfo.useableStart + flashInfo.useableLength) {
    return DfIoStatusError(ADMXRC3_INVALID_REGION);
  }

  ticket.count = 1;
  ticket.resources[0].code = CORE_RESOURCE_FLASH;
  ticket.resources[0].index = (uint8_t)bankIndex;
  acqStatus = pDevCtx->coreInterface.pAcquireInitialize(pDevCtx->pCoreContext, &ticket);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }
  acqStatus = pDevCtx->coreInterface.pAcquireSync(pDevCtx->pCoreContext, &ticket, 0);
  if (CoreAcquireSuccess != acqStatus) {
    return mapAcquireStatus(acqStatus);
  }

  dfRequestSetContext(pReq, &context);
  dfEventInit(&context.event);
  context.bCancel = FALSE;
  if (!dfRequestSetCallback(pReq, cancelCallback)) {
    dfEventWait(&context.event);
    status = context.reason;
    goto done2;
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuf, length, FALSE);

  pStagingBuffer = pDevCtx->flash[bankIndex].pStagingBuffer;
  while (length) {
    flashStatus = pDevCtx->coreInterface.pGetFlashBlockInfo(pDevCtx->pCoreContext, bankIndex, offset, &blockInfo);
    if (CoreFlashSuccess != flashStatus) {
      status = mapFlashStatus(flashStatus);
      goto done;
    }
    blockOffset = offset - blockInfo.address;
    chunk = blockInfo.length - blockOffset;
    if (chunk > length) {
      chunk = length;
    }
    dfCopyUK(pStagingBuffer, pUserBuf, (size_t)chunk); /* Cast safe because (chunk <= length) and length is of type 'size_t' */
    /* Cast in following line is safe because (chunk <= length) and length is of type 'size_t' */
    flashStatus = pDevCtx->coreInterface.pFlashWrite(pDevCtx->pCoreContext, &ticket, bankIndex, offset, (size_t)chunk, pStagingBuffer);
    if (CoreFlashSuccess != flashStatus) {
      status = mapFlashStatus(flashStatus);
      goto done;
    }
    offset += chunk;
    pUserBuf += chunk;
    length -= (size_t)chunk;
    if (context.bCancel) {
      break;
    }
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

done:
  if (!dfRequestClearCallback(pReq)) {
    dfEventWait(&context.event);
    status = context.reason;
  }

done2:
  pDevCtx->coreInterface.pRelease(pDevCtx->pCoreContext, &ticket);
  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteFlashThunk(
  Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_WRITEFLASH* pIoctl32 = (IOCTLS32_ADMXRC3_WRITEFLASH*)pBuffer;
  IOCTLS_ADMXRC3_WRITEFLASH ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.flashIndex = pIoctl32->in.flashIndex;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.offset = pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.pBuffer = dfThunkConstPtr(pIoctl32->in.pBuffer);
  return ioctlWriteFlash(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif
