/*
** File: model_admxrc2.c
** Project: ADMXRC3 module driver
** Purpose: ADM-XRC-II specific code, including the function to bootstrap
**          the device context.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>

#include <core/admxrc2.h>
#include "device.h"
#include "info.h"
#include "model_boot.h"

static ADMXRC3_STATUS
getBankInfo(
  Admxrc3DeviceContext* pDevCtx,
  unsigned int bankIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETBANKINFO* pIoctl)
{
  VpdAdmxrc2Rev0* pVpd = (VpdAdmxrc2Rev0*)pVpdBuffer;
  uint32_t zbt, bank;

  if (bankIndex >= 6) {
    return ADMXRC3_INVALID_INDEX;
  }
  zbt = pVpd->zbt01 | ((uint32_t)pVpd->zbt23 << 8) | ((uint32_t)pVpd->zbt45 << 16);
  bank = (zbt >> (bankIndex * 4)) & 0xfU;
  pIoctl->out.maximumFrequency = 0;
  pIoctl->out.minimumFrequency = 0;
  pIoctl->out.physicalSize = (uint64_t)0x1U << ((bank >> 2) + 18); 
  pIoctl->out.physicalDataWidth = 32;
  if (pVpd->fpgaType == ADMXRC3_FPGA_2V3000) {
    pIoctl->out.physicalEccWidth = 0;
  } else {
    pIoctl->out.physicalEccWidth = 4;
  }
  pIoctl->out.typeMask = (zbt >> (bankIndex * 4)) & 0x3U;
  pIoctl->out.connectivityMask = 0x1U;
  pIoctl->out.bPresent = ((zbt >> (bankIndex * 4)) & 0x3U) ? TRUE : FALSE;

  return ADMXRC3_SUCCESS;
}

static void
getCardInfo(
	Admxrc3DeviceContext* pDevCtx,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETCARDINFO* pIoctl)
{
  VpdAdmxrc2Rev0* pVpd = (VpdAdmxrc2Rev0*)pVpdBuffer;
  uint32_t zbtAll;
  unsigned int i;

  pIoctl->out.serialNumber = pVpd->serialNumber;
  pIoctl->out.numMemoryBank = 6;
  pIoctl->out.memoryBankPresent = 0;
  zbtAll = pVpd->zbt01 | ((uint32_t)pVpd->zbt23 << 8) | ((uint32_t)pVpd->zbt45 << 16);
  for (i = 0; i < 6; i++) {
    uint8_t bank = (uint8_t)(zbtAll >> (i * 4)) & 0xfU;

    if (0 == (bank & 0x3)) {
      /* Bank is not fitted */
    } else {
      /* Bank is fitted */
      pIoctl->out.memoryBankPresent |= (0x1U << i);
    }
  }
}

static ADMXRC3_STATUS
getFpgaInfo(
	Admxrc3DeviceContext* pDevCtx,
  unsigned int targetIndex,
  const void* pVpdBuffer,
  IOCTLS_ADMXRC3_GETFPGAINFO* pIoctl)
{
  VpdAdmxrc2Rev0* pVpd = (VpdAdmxrc2Rev0*)pVpdBuffer;

  if (0 != targetIndex) {
    return ADMXRC3_INVALID_INDEX;
  }

  pIoctl->out.flags = 0;
  setupIoctlsGetFpgaInfo(pDevCtx, pIoctl, pVpd->fpgaType, ADMXRC3_PACKAGE_MAKEFF(1152), NULL, NULL, TRUE);

  return ADMXRC3_SUCCESS;
}

void
admxrc3BootstrapAdmxrc2(
	Admxrc3DeviceContext* pDevCtx)
{
  dfDebugPrint(1, ("admxrc3BootstrapAdmxrc2: entered\n"));

  pDevCtx->methods.pGetBankInfo = getBankInfo;
  pDevCtx->methods.pGetCardInfo = getCardInfo;
  pDevCtx->methods.pGetFpgaInfo = getFpgaInfo;
  pDevCtx->bReverseSelectMapData = TRUE;
}
