/*
** File: pio.c  
** Project: ADMXRC3 module driver
** Purpose: OS-independent IOCTL handlers for programmed I/O.
**
** (C) Copyright Alpha Data 2009
*/

#include <df.h>
#include <core/coreif.h>
#include "device.h"
#include "pio.h"
#include <admxrc3/types.h>

/* Length threshold at which we install a cancel callback for I/O requests */
#define CANCEL_THRESHOLD (0x1000U)

typedef struct _Context {
  boolean_t bCancel;
  DfIoStatus reason;
  DfEvent event;
} Context;

static const size_t maxChunk = 1024;

static void
cancelCallback(
  DfClientRequest* pReq,
  DfIoStatus reason)
{
  Context* pContext = (Context*)dfRequestGetContext(pReq);

  pContext->reason = reason;
  pContext->bCancel = TRUE;
  dfEventSignal(&pContext->event);
}

DfIoStatus
ioctlRead(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = 0xffU | ADMXRC3_ACCESS_CONSTANTBIT;
#endif
  IOCTLS_ADMXRC3_READ* pIoctl = (IOCTLS_ADMXRC3_READ*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreWindowAddressStatus windowStatus;
  Context context;
  unsigned int windowIndex;
  uint32_t flags;
  size_t offset;
  size_t length, chunk, remaining;
  uint8_t* pUserBuf;
  uint8_t* pKernelAddr;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  windowIndex = pIoctl->in.windowIndex;
  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  pUserBuf = pIoctl->in.pBuffer;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  windowStatus = pDevCtx->coreInterface.pGetWindowAddress(pDevCtx->pCoreContext, windowIndex, offset, length, (void**)&pKernelAddr);
  switch (windowStatus) {
  case CoreWindowAddressSuccess:
    break;
  case CoreWindowAddressInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX); 
  case CoreWindowAddressInvalidRegion:
    return DfIoStatusError(ADMXRC3_INVALID_REGION); 
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR); 
  }

  context.bCancel = FALSE;
  if (length >= CANCEL_THRESHOLD) {
    dfRequestSetContext(pReq, &context);
    dfEventInit(&context.event);
    if (!dfRequestSetCallback(pReq, cancelCallback)) {
      return DfIoStatusError(ADMXRC3_CANCELLED); 
    }
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuf, length, TRUE);

  remaining = length;
  while (remaining) {
    chunk = (remaining > maxChunk) ? maxChunk : remaining;
    dfCopyIU(pUserBuf, pKernelAddr, chunk);
    pUserBuf += chunk;
    pKernelAddr += chunk;
    remaining -= chunk;
    if (context.bCancel) {
      break;
    }
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

  if (length >= CANCEL_THRESHOLD && !dfRequestClearCallback(pReq)) {
    dfEventWait(&context.event);
    status = context.reason;
  }

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlReadThunk(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_READ* pIoctl32 = (IOCTLS32_ADMXRC3_READ*)pBuffer;
  IOCTLS_ADMXRC3_READ ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.windowIndex = pIoctl32->in.windowIndex;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.offset = (size_t)pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.pBuffer = dfThunkPtr(pIoctl32->in.pBuffer);
  return ioctlRead(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif

DfIoStatus
ioctlWrite(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
#if defined(ADMXRC3_CHECK_FLAGS)
  static const uint32_t legalFlagsMask = 0xffU | ADMXRC3_ACCESS_CONSTANTBIT;
#endif
  IOCTLS_ADMXRC3_WRITE* pIoctl = (IOCTLS_ADMXRC3_WRITE*)pBuffer;
  DfIoStatus status = DfIoStatusSuccess;
  CoreWindowAddressStatus windowStatus;
  Context context;
  unsigned int windowIndex;
  uint32_t flags;
  size_t offset;
  size_t length, chunk, remaining;
  const uint8_t* pUserBuf;
  uint8_t* pKernelAddr;

  if (NULL == pIoctl || sizeof(pIoctl->in) != inSize) {
    return DfIoStatusInvalid;
  }
  if (!pClCtx->bPrivileged) {
    return DfIoStatusError(ADMXRC3_ACCESS_DENIED);
  }

  windowIndex = pIoctl->in.windowIndex;
  flags = pIoctl->in.flags;
  offset = pIoctl->in.offset;
  length = pIoctl->in.length;
  pUserBuf = (const uint8_t*)pIoctl->in.pBuffer;
#if defined(ADMXRC3_CHECK_FLAGS)
  if (0 != (flags & ~legalFlagsMask)) {
    return DfIoStatusError(ADMXRC3_INVALID_FLAG);
  }
#endif

  dfDebugPrint(6, ("ioctlWrite: windowIndex=%u flags=0x%lx offset=%p length=%p pUserBuf=%p\n",
    windowIndex, (unsigned long)flags, (uintptr_t)offset, (uintptr_t)length, pUserBuf));

  windowStatus = pDevCtx->coreInterface.pGetWindowAddress(pDevCtx->pCoreContext, windowIndex, offset, length, (void**)&pKernelAddr);
  switch (windowStatus) {
  case CoreWindowAddressSuccess:
    break;
  case CoreWindowAddressInvalidIndex:
    return DfIoStatusError(ADMXRC3_INVALID_INDEX); 
  case CoreWindowAddressInvalidRegion:
    return DfIoStatusError(ADMXRC3_INVALID_REGION); 
  default:
    return DfIoStatusError(ADMXRC3_UNEXPECTED_ERROR); 
  }

  context.bCancel = FALSE;
  if (length >= CANCEL_THRESHOLD) {
    dfRequestSetContext(pReq, &context);
    dfEventInit(&context.event);
    if (!dfRequestSetCallback(pReq, cancelCallback)) {
      return DfIoStatusError(ADMXRC3_CANCELLED); 
    }
  }

  DF_USER_SPACE_ACCESS_BEGIN

  dfUserBufferValidate(pUserBuf, length, FALSE);

  remaining = length;
  while (remaining) {
    chunk = (remaining > maxChunk) ? maxChunk : remaining;
    dfCopyUI(pKernelAddr, pUserBuf, chunk);
    pUserBuf += chunk;
    pKernelAddr += chunk;
    remaining -= chunk;
    if (context.bCancel) {
      break;
    }
  }

  DF_USER_SPACE_ACCESS_END({
    status = DfIoStatusError(ADMXRC3_INVALID_BUFFER);
  })

  if (length >= CANCEL_THRESHOLD && !dfRequestClearCallback(pReq)) {
    dfEventWait(&context.event);
    status = context.reason;
  }

  return status;
}

#if DF_NEED_THUNK
DfIoStatus
ioctlWriteThunk(
	Admxrc3DeviceContext* pDevCtx,
	Admxrc3ClientContext* pClCtx,
  DfClientRequest* pReq,
  void* pBuffer,
  unsigned int inSize)
{
  IOCTLS32_ADMXRC3_WRITE* pIoctl32 = (IOCTLS32_ADMXRC3_WRITE*)pBuffer;
  IOCTLS_ADMXRC3_WRITE ioctl;

  if (NULL == pIoctl32 || sizeof(pIoctl32->in) != inSize) {
    return DfIoStatusInvalid;
  }

  ioctl.in.windowIndex = pIoctl32->in.windowIndex;
  ioctl.in.flags = pIoctl32->in.flags;
  ioctl.in.offset = (size_t)pIoctl32->in.offset;
  ioctl.in.length = (size_t)pIoctl32->in.length;
  ioctl.in.pBuffer = dfThunkConstPtr(pIoctl32->in.pBuffer);
  return ioctlWrite(pDevCtx, pClCtx, pReq, &ioctl, sizeof(ioctl.in));
}
#endif
